class View:
    def activate_view(self):
        """
        Run every time the view is activated by code or by clicking the tool button associated with the view.
        This is useful if the view wants to reset to a default state every time the tool button is clicked.
        """
        pass

    def can_disable_view(self) -> bool:
        """
        Can the view be disabled.
        Return False if for some reason the view cannot be disabled.

        :return: True if the view can be disabled. False otherwise.
        """
        return True

    def disable_view(self):
        """
        This is run when a view is about to be removed from a view container.
        This gives the view a chance to do any cleaning up before it is removed and potentially destroyed.
        This method must leave the view in a state where it can be destroyed or activated again.
        """
        pass



from __future__ import annotations
import os
from typing import Optional

from PySide6.QtCore import QLocale, QCoreApplication

from amulet_editor.models.localisation import ATranslator

import amulet_team_locale

import amulet_team_3d_viewer


# Qt only weekly references this. We must hold a strong reference to stop it getting garbage collected
_translator: Optional[ATranslator] = None


def load_plugin():
    global _translator
    _translator = ATranslator()
    _locale_changed()
    QCoreApplication.installTranslator(_translator)
    amulet_team_locale.locale_changed.connect(_locale_changed)


def _locale_changed():
    _translator.load_lang(
        QLocale(),
        "",
        directory=os.path.join(*amulet_team_3d_viewer.__path__, "_resources", "lang"),
    )


def unload_plugin():
    QCoreApplication.removeTranslator(_translator)



from amulet_editor import __version__

from PySide6.QtCore import Qt
from PySide6.QtGui import QImage, QPixmap
from amulet_editor.data import build
from ._home import Ui_HomePage


class HomePage(Ui_HomePage):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        amulet_logo = QPixmap(QImage(build.get_resource("images/amulet_logo.png")))
        amulet_logo = amulet_logo.scaledToHeight(128)
        self.lbl_app_icon.setPixmap(amulet_logo)
        self.lbl_app_version.setText(f"Version {__version__}")
        self._central_layout.setAlignment(Qt.AlignCenter)
        self.cbo_language.addItem("English")



from __future__ import annotations

import json
import os.path
from typing import Optional, Type, Protocol
from abc import ABC, abstractmethod
from packaging.version import Version
from packaging.specifiers import SpecifierSet

from amulet_editor.data.paths._plugin import first_party_plugin_directory
from ._data import PluginData, PluginDataDepends
from ._state import PluginState
from ._requirement import Requirement
from ._uid import LibraryUID


_plugin_classes: dict[int, Type[PluginContainer]] = {}


class Plugin(Protocol):
    def load_plugin(self):
        """
        Logic run when the plugin is started.
        All dependencies will be started when this is called.
        Plugins may implement this method but must not call it.
        """
        ...

    def unload_plugin(self):
        """
        Logic run when the plugin is stopped.
        Dependents will be stopped at this point but dependencies are not.
        This must leave the program in the same state as it was before load_plugin was called.
        Plugins may implement this method but must not call it.
        """
        ...


class PluginContainer(ABC):
    data: PluginData
    instance: Optional[Plugin]  # The instance of the plugin.
    state: PluginState

    FormatVersion: int = None

    def __init__(self, data: PluginData):
        self.data = data
        self.instance = None
        self.state = PluginState.Disabled

    @classmethod
    def from_path(cls, plugin_path: str) -> PluginContainer:
        """
        Populate a PluginContainer instance from the data in a directory.
        plugin_path is the system path to a directory containing a plugin.json file.
        """
        with open(os.path.join(plugin_path, "plugin.json")) as f:
            plugin_data = json.load(f)
        if not isinstance(plugin_data, dict):
            raise TypeError("plugin.json must be a dictionary.")

        # Get the plugin identifier
        format_version = plugin_data.get("format_version", 1)
        if not isinstance(format_version, int):
            raise TypeError("plugin.json[format_version] must be an int")
        try:
            cls2 = _plugin_classes[format_version]
        except KeyError:
            raise ValueError(f"Unknown plugin format version {format_version}")
        else:
            return cls2.from_data(
                plugin_path,
                plugin_data,
                os.path.dirname(plugin_path) == first_party_plugin_directory(),
            )

    @classmethod
    @abstractmethod
    def from_data(
        cls, plugin_path: str, plugin_data: dict, first_party: bool
    ) -> PluginContainer:
        raise NotImplementedError

    def __init_subclass__(cls, **kwargs):
        if cls.FormatVersion is None:
            raise NotImplementedError("FormatVersion has not been set.")
        if cls.FormatVersion in _plugin_classes:
            raise ValueError(
                f"Two classes have been registered with format version {cls.FormatVersion}"
            )
        _plugin_classes[cls.FormatVersion] = cls


class PluginContainerV1(PluginContainer):
    FormatVersion = 1

    def __repr__(self):
        return f"PluginContainerV1({self.data.path})"

    @classmethod
    def from_data(
        cls, plugin_path: str, plugin_data: dict, first_party: bool
    ) -> PluginContainerV1:
        """
        Populate a PluginContainer instance from the data in a directory.
        plugin_path is the system path to a directory containing a plugin.json file.
        """
        # Get the plugin identifier
        plugin_identifier = plugin_data.get("identifier")
        if not isinstance(plugin_identifier, str):
            raise TypeError("plugin.json[identifier] must be a string")
        if not plugin_identifier.isidentifier():
            raise ValueError(
                "plugin.json[identifier] must be a valid python identifier"
            )

        # Get the plugin version
        plugin_version_string = plugin_data.get("version")
        if not isinstance(plugin_version_string, str):
            raise TypeError("plugin.json[version] must be a PEP 440 version string")
        plugin_version = Version(plugin_version_string)

        # Get the plugin name
        plugin_name = plugin_data.get("name")
        if not isinstance(plugin_version_string, str):
            raise TypeError("plugin.json[name] must be a string")

        # Get the plugin dependencies
        depends_raw: dict = plugin_data.get("depends")
        if not isinstance(depends_raw, dict):
            raise TypeError(
                'plugin.json[depends] must be a dictionary of the form {"python": "~=3.9", "library": [], "plugin": []}'
            )

        python_raw = depends_raw.get("python")
        if not isinstance(python_raw, str):
            raise TypeError(
                'plugin.json[depends][python] must be a string of the form "~=3.9"'
            )
        python = SpecifierSet(python_raw)

        library_depends_raw: dict = depends_raw.get("library", [])
        if not isinstance(library_depends_raw, list) and all(
            isinstance(d, str) for d in library_depends_raw
        ):
            raise TypeError(
                'plugin.json[depends][library] must be a list of string identifiers and version specifiers if defined.\nEg. ["PySide6_Essentials~=6.4"]'
            )
        library_depends = tuple(map(Requirement.from_string, library_depends_raw))

        plugin_depends_raw: dict = depends_raw.get("plugin", [])
        if not isinstance(plugin_depends_raw, list) and all(
            isinstance(d, str) for d in plugin_depends_raw
        ):
            raise TypeError(
                'plugin.json[depends][plugin] must be a list of string identifiers and version specifiers if defined.\nEg. ["plugin_1~=1.0", "plugin_2~=1.3"]'
            )
        plugin_depends = tuple(map(Requirement.from_string, plugin_depends_raw))

        # Get the locked state
        locked = bool(first_party and plugin_data.get("locked"))

        return cls(
            PluginData(
                LibraryUID(plugin_identifier, plugin_version),
                plugin_path,
                plugin_name,
                PluginDataDepends(
                    python,
                    library_depends,
                    plugin_depends,
                ),
                locked,
            )
        )



from typing import Optional

from amulet_editor.tools.project._components import MCFunctionHighlighter, QCodeEditor
from PySide6.QtGui import QFontDatabase, QSyntaxHighlighter
from PySide6.QtWidgets import QVBoxLayout, QWidget


class ProgramsPage(QWidget):
    def __init__(self, parent: Optional[QWidget] = None):
        if parent is None:
            super().__init__()
        else:
            super().__init__(parent=parent)

        self.highlighter: Optional[QSyntaxHighlighter] = None

        self.setupUi()

    def setupUi(self):
        font = QFontDatabase.systemFont(QFontDatabase.SystemFont.FixedFont)

        self.txe_file = QCodeEditor()
        self.txe_file.setFont(font)

        self.hlt_mcfunction = MCFunctionHighlighter()
        self.hlt_mcfunction.setDocument(self.txe_file.document())
        self.hlt_mcfunction.setDocument(None)

        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.txe_file)

        self.setLayout(layout)



from __future__ import annotations
import logging
from math import sin, cos, radians

from PySide6.QtCore import Qt, QPoint, Slot
from PySide6.QtGui import (
    QOpenGLFunctions,
    QOpenGLContext,
    QMouseEvent,
    QShowEvent,
    QHideEvent,
    QWheelEvent,
    QCursor,
    QGuiApplication,
)
from PySide6.QtOpenGLWidgets import QOpenGLWidget

from OpenGL.GL import (
    GL_COLOR_BUFFER_BIT,
    GL_DEPTH_BUFFER_BIT,
    GL_DEPTH_TEST,
    GL_CULL_FACE,
)

from amulet_editor.data.project import get_level

from amulet_team_resource_pack._api import get_resource_pack_container

from ._camera import Camera, Location, Rotation
from ._key_catcher import KeySrc, KeyCatcher
from ._level_geometry import WidgetLevelGeometry
from ._resource_pack import get_gl_resource_pack_container

log = logging.getLogger(__name__)


"""
GPU Memory Deallocation
GPU memory must be associated either with the context or the widget.
Widget memory must be destroyed when the widget is destroyed.
    self.destroyed.connect(func)
Context memory must be destroyed when the context is destroyed.
    self.context().aboutToBeDestroyed.connect(func)
    
Note that if you use a method bound to the instance being destroyed IT WILL NOT BE CALLED.
Due to this all references to GPU memory must be stored in a class stored as an attribute
and a method on that instance should be bound to one of the above examples.
"""


class FirstPersonCanvas(QOpenGLWidget, QOpenGLFunctions):
    background_colour = (0.61, 0.70, 0.85)

    def __init__(self, parent=None):
        QOpenGLWidget.__init__(self, parent)
        QOpenGLFunctions.__init__(self)

        self._camera = Camera()
        self.camera.transform_changed.connect(self.update)
        self.camera.location = Location(0, 0, 0)
        self._start_pos = QPoint()
        self._mouse_captured = False

        self._speed = 1

        self._key_catcher = KeyCatcher()
        self.installEventFilter(self._key_catcher)
        self._key_catcher.connect_repeating(
            self._forwards, (KeySrc.Keyboard, Qt.Key.Key_I), frozenset(), 10
        )
        self._key_catcher.connect_repeating(
            self._right, (KeySrc.Keyboard, Qt.Key.Key_L), frozenset(), 10
        )
        self._key_catcher.connect_repeating(
            self._backwards, (KeySrc.Keyboard, Qt.Key.Key_K), frozenset(), 10
        )
        self._key_catcher.connect_repeating(
            self._left, (KeySrc.Keyboard, Qt.Key.Key_J), frozenset(), 10
        )
        self._key_catcher.connect_repeating(
            self._up, (KeySrc.Keyboard, Qt.Key.Key_Space), frozenset(), 10
        )
        self._key_catcher.connect_repeating(
            self._down, (KeySrc.Keyboard, Qt.Key.Key_Semicolon), frozenset(), 10
        )

        self.camera.location_changed.connect(self._on_move)

        level = get_level()
        self._render_level = WidgetLevelGeometry(level)
        self._render_level.geometry_changed.connect(self.update)
        self._render_level.set_dimension(level.dimensions[0])
        self._render_level.set_location(0, 0)

        self._resource_pack_container = get_resource_pack_container(level)
        self._resource_pack_container.changing.connect(
            lambda prom: prom.progress_change.connect(
                lambda prog: print(f"Loading resource pack {prog}")
            )
        )
        self._gl_resource_pack_container = get_gl_resource_pack_container(level)
        self._gl_resource_pack_container.changing.connect(
            lambda prom: prom.progress_change.connect(
                lambda prog: print(f"Loading GL resource pack {prog}")
            )
        )
        self._resource_pack_container.init()

    def initializeGL(self):
        """Private initialisation method called by the QOpenGLWidget"""
        log.debug(f"Initialising GL for {self}")
        self.initializeOpenGLFunctions()
        self.glClearColor(*self.background_colour, 1)
        self._render_level.initializeGL()

    @property
    def camera(self) -> Camera:
        return self._camera

    def showEvent(self, event: QShowEvent) -> None:
        self._render_level.start()

    def hideEvent(self, event: QHideEvent) -> None:
        self._render_level.stop()

    def paintGL(self):
        """Private paint method called by the QOpenGLWidget"""
        if QOpenGLContext.currentContext() is not self.context():
            log.error("Tried to paint from a different context.")
            return
        self.glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        self.glEnable(GL_DEPTH_TEST)

        self._render_level.paintGL(
            self.camera.intrinsic_matrix, self.camera.extrinsic_matrix
        )

    def resizeGL(self, width, height):
        """Private resize method called by the QOpenGLWidget"""
        self.camera.set_perspective_projection(45, width / height, 0.01, 10_000)

    def mousePressEvent(self, event: QMouseEvent):
        if event.buttons() & Qt.MouseButton.RightButton:
            self._mouse_captured = True
            self._start_pos = event.globalPos()
            self.setFocus()
            QGuiApplication.setOverrideCursor(QCursor(Qt.CursorShape.BlankCursor))

    def mouseMoveEvent(self, event: QMouseEvent):
        if event.buttons() & Qt.MouseButton.RightButton:
            pos = event.globalPos()
            dx = pos.x() - self._start_pos.x()
            dy = pos.y() - self._start_pos.y()

            azimuth, elevation = self.camera.rotation
            azimuth += dx / 8
            elevation += dy / 8
            self.camera.rotation = Rotation(azimuth, elevation)

            QCursor.setPos(self._start_pos)
            # On some systems setPos does not work. We must reset _start_pos to the new pos
            self._start_pos = QCursor.pos()

    def mouseReleaseEvent(self, event: QMouseEvent):
        if self._mouse_captured:
            self._mouse_captured = False
            QGuiApplication.restoreOverrideCursor()

    def wheelEvent(self, event: QWheelEvent) -> None:
        if event.angleDelta().y() > 0:
            self._faster()
        else:
            self._slower()

    def _on_move(self):
        x, _, z = self.camera.location
        self._render_level.set_location(x // 16, z // 16)

    def _move_relative(self, angle: int, dt: float):
        x, y, z = self.camera.location
        azimuth = radians(self.camera.rotation.azimuth + angle)
        self.camera.location = Location(
            x - sin(azimuth) * self._speed * dt, y, z + cos(azimuth) * self._speed * dt
        )

    @Slot()
    def _forwards(self, dt: float):
        self._move_relative(180, dt)

    @Slot()
    def _right(self, dt: float):
        self._move_relative(270, dt)

    @Slot()
    def _backwards(self, dt: float):
        self._move_relative(0, dt)

    @Slot()
    def _left(self, dt: float):
        self._move_relative(90, dt)

    @Slot()
    def _up(self, dt: float):
        x, y, z = self.camera.location
        self.camera.location = Location(x, y + self._speed * dt, z)

    @Slot()
    def _down(self, dt: float):
        x, y, z = self.camera.location
        self.camera.location = Location(x, y - self._speed * dt, z)

    @Slot()
    def _faster(self):
        self._speed *= 1.1

    @Slot()
    def _slower(self):
        self._speed /= 1.1



"""
This module manages resource pack objects for each level
"""

from weakref import WeakKeyDictionary
from threading import Lock, RLock
from typing import Optional
import logging

from PySide6.QtCore import QObject, QCoreApplication, Signal

from amulet.api.level import BaseLevel
from amulet_editor.models.generic._promise import Promise
from amulet_editor.models.widgets.traceback_dialog import DisplayException

from minecraft_model_reader import BaseResourcePackManager
from minecraft_model_reader.api.resource_pack import (
    load_resource_pack_manager,
)
from minecraft_model_reader.api.resource_pack.java.download_resources import (
    get_java_vanilla_latest_iter,
    get_java_vanilla_fix,
)

log = logging.getLogger(__name__)


class ResourcePackContainer(QObject):
    # Emitted with a promise object when a resource pack change is started.
    changing = Signal(object)  # Promise[None]
    # Emitted when the resource pack has changed.
    changed = Signal()

    def __init__(self):
        super().__init__()
        self._lock = RLock()
        self._resource_pack: Optional[BaseResourcePackManager] = None
        self._loader: Optional[Promise[None]] = None

    @property
    def loader(self) -> Optional[Promise[None]]:
        """
        If the resource pack is being loaded this will be a promise.
        This can be used to update a GUI to show the progress.
        :return: A Promise instance if a resource pack is being loaded otherwise None.
        """
        return self._loader

    @property
    def loaded(self) -> bool:
        """
        Is there a valid resource pack.
        :return: True if loaded otherwise False.
        """
        return self._resource_pack is not None

    @property
    def resource_pack(self) -> BaseResourcePackManager:
        """
        :return: The active resource pack for this level.
        :raises RuntimeError: If the resource pack has not been initialised yet.
        """
        rp = self._resource_pack
        if rp is None:
            raise RuntimeError(
                "The ResourcePackManager for this level has not been loaded yet."
            )
        return rp

    def init(self):
        """
        Initialise the default resource pack for this level if one does not already exist.
        This is completed asynchronously. A promise is emitted from changing.

        >>> def load(level: BaseLevel):
        >>>     container = get_resource_pack_container(level)
        >>>     promise = container.init(level)
        >>>     def on_ready():
        >>>         print("finished")
        >>>         if promise.get_return():
        >>>             print("resource pack was already loaded")
        >>>         else:
        >>>             print("resource pack has been loaded.")
        >>>     promise.ready.connect(on_ready)
        >>>     promise.progress_change.connect(lambda progress: print(progress))
        >>>     promise.start()
        """

        def init(promise_data: Promise.Data) -> bool:
            with self._lock, DisplayException("Error initialising the resource pack."):
                if self._resource_pack is None:
                    # TODO: support other resource pack formats
                    promise_data.progress_text_change.emit(
                        QCoreApplication.translate(
                            "ResourcePack", "downloading_resource_pack", None
                        )
                    )
                    try:
                        it = get_java_vanilla_latest_iter()
                        while True:
                            progress = next(it)
                            promise_data.progress_change.emit(progress * 0.5)
                            if promise_data.is_cancel_requested():
                                raise Promise.OperationCanceled()
                    except StopIteration as e:
                        vanilla = e.value

                    self._resource_pack = load_resource_pack_manager(
                        [vanilla, get_java_vanilla_fix()], load=False
                    )
                    promise_data.progress_text_change.emit(
                        QCoreApplication.translate(
                            "ResourcePack", "loading_resource_pack", None
                        )
                    )
                    for progress in self._resource_pack.reload():
                        promise_data.progress_change.emit(0.5 + progress * 0.5)
                    self.changed.emit()
                    log.debug("Loaded resource pack.")
                    return False
                return True

        promise_ = Promise[bool](init)
        old_loader = self._loader
        if old_loader is not None:
            old_loader.cancel()
        self._loader = promise_
        self.changing.emit(promise_)
        promise_.start()

    # def set_resource_pack(self, resource_pack: BaseResourcePackManager):
    #     """
    #     Set the resource pack.
    #     Will emit a signal from changed after setting
    #
    #     :param resource_pack: The resource pack to set.
    #     """
    #     if not isinstance(resource_pack, BaseResourcePackManager):
    #         raise TypeError("resource_pack must be an instance of BaseResourcePackManager")
    #     with self._lock:
    #         self._resource_pack = resource_pack
    #         self.changed.emit()


_lock = Lock()
_level_data: WeakKeyDictionary[BaseLevel, ResourcePackContainer] = WeakKeyDictionary()


def get_resource_pack_container(level: BaseLevel) -> ResourcePackContainer:
    with _lock:
        if level not in _level_data:
            _level_data[level] = ResourcePackContainer()
        return _level_data[level]



from typing import Optional

from amulet_editor.data import build
from amulet_editor.models.widgets._label import QHoverLabel
from PySide6.QtCore import QEvent, QSize, Qt
from PySide6.QtGui import QColor, QEnterEvent, QPainter, QPaintEvent, QImage
from PySide6.QtWidgets import QWidget, QStyleOption, QStyle, QVBoxLayout, QPushButton
from PySide6.QtSvgWidgets import QSvgWidget


class AStylableSvgWidget(QSvgWidget):
    """
    A subclass of QSvgWidget that adds support for colouring via style sheets.
    Set the background via QSS and this widget will merge the colour of the background with the transparency of the SVG icon.
    """

    def paintEvent(self, event: QPaintEvent):
        buffer_width = max(self.width(), 50)
        buffer_height = max(self.height(), 50)

        buffer = QImage(buffer_width, buffer_height, QImage.Format.Format_ARGB32)
        buffer.fill(QColor(0, 0, 0, 0))  # Fill with transparency

        # Init the painter
        painter = QPainter(buffer)

        # Draw the SVG
        self.renderer().render(painter)
        # If the buffer is larger than the widget apply scaling
        painter.scale(buffer_width / self.width(), buffer_height / self.height())

        # Use the alpha current alpha with the future colour
        painter.setCompositionMode(QPainter.CompositionMode.CompositionMode_SourceIn)
        # Draw the normal background
        opt = QStyleOption()
        opt.initFrom(self)
        self.style().drawPrimitive(
            QStyle.PrimitiveElement.PE_Widget, opt, painter, self
        )

        # Finish painting
        painter.end()

        # Draw the image to the widget.
        painter = QPainter(self)
        painter.drawImage(self.rect(), buffer)
        painter.end()


class AIconButton(QPushButton):
    """A QPushButton containing a stylable icon."""

    def __init__(self, icon_name: str = "question-mark.svg", parent: QWidget = None):
        super().__init__(parent)
        self.setProperty("hover", "false")
        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self._icon = AStylableSvgWidget(build.get_resource(f"icons/tabler/{icon_name}"))
        self._layout.addWidget(self._icon)

    def setIcon(self, icon_name: Optional[str] = None):
        self._icon.load(build.get_resource(f"icons/tabler/{icon_name}"))

    def setIconSize(self, size: QSize):
        self._icon.setFixedSize(size)

    def enterEvent(self, event: QEnterEvent):
        self.setProperty("hover", "true")
        self.setStyleSheet("/* /")  # Force a style update.
        super().enterEvent(event)

    def leaveEvent(self, event: QEvent):
        if not self.isChecked():
            self.setProperty("hover", "false")
        self.setStyleSheet("/* /")  # Force a style update.
        super().leaveEvent(event)


class ATooltipIconButton(AIconButton):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._hlbl_tooltip: Optional[QHoverLabel] = None

    def enterEvent(self, event: QEnterEvent):
        if self._hlbl_tooltip is not None and len(self._hlbl_tooltip.text()) > 0:
            self._hlbl_tooltip.show()
        super().enterEvent(event)

    def leaveEvent(self, event: QEvent):
        if self._hlbl_tooltip is not None:
            self._hlbl_tooltip.hide()
        super().leaveEvent(event)

    def toolTip(self) -> str:
        return "" if self._hlbl_tooltip is None else self._hlbl_tooltip.text()

    def setToolTip(self, label: str):
        if self._hlbl_tooltip is None:
            self._hlbl_tooltip = QHoverLabel(label, self)
            self._hlbl_tooltip.hide()
        else:
            self._hlbl_tooltip.setText(label)



from __future__ import annotations

import inspect
import time
from typing import NamedTuple, Optional
from types import FrameType
from threading import RLock
import os
from os.path import normpath, samefile
import glob
import logging
from importlib import import_module
from importlib.util import spec_from_file_location, module_from_spec
from importlib.metadata import version, packages_distributions
from queue import Queue, Empty
from enum import Enum
import sys
from collections import UserDict
import re
import traceback
import builtins

from packaging.version import Version

from PySide6.QtCore import QThread, Signal, QObject

from amulet_editor.application._splash import Splash
from amulet_editor.application._invoke import invoke

from amulet_editor.data.paths._plugin import (
    first_party_plugin_directory,
    third_party_plugin_directory,
)

# from amulet_editor.data.process._messaging import (
#     register_global_function,
#     call_in_parent,
#     call_in_children,
# )
from amulet_editor.models.plugin import LibraryUID
from amulet_editor.models.plugin._state import PluginState
from amulet_editor.models.plugin._container import PluginContainer
from amulet_editor.models.plugin._requirement import Requirement
from amulet_editor.models.widgets.traceback_dialog import display_exception


log = logging.getLogger(__name__)
PythonVersion = Version(".".join(map(str, sys.version_info[:3])))
Packages = packages_distributions()

"""
Notes:
First party plugins are stored in amulet_editor.plugins
They cannot be imported from that path because the __init__.py disallows it.
Third party plugins are imported as a zip and extracted to a writable directory with a UUID as the name.
Custom code loads the plugin package into sys.modules under its package name. Adding amulet_editor.plugins as sources root helps the IDE understand this.
TODO: look into generating stub files for the active plugins to help with development on the compiled version.
Plugins can import directly from other plugins to access static classes and functions 
"""


def plugin_dirs() -> tuple[str, str]:
    return first_party_plugin_directory(), third_party_plugin_directory()


class PluginJobType(Enum):
    Enable = 1
    Disable = 2
    Reload = 3


class PluginJob(NamedTuple):
    plugin_identifier: LibraryUID
    job_type: PluginJobType


# A lock for the plugin data. Code must acquire this before touching the plugin data
_plugin_lock = RLock()

# The plugin data
_plugins: dict[LibraryUID, PluginContainer] = {}

# A queue of jobs to apply to the plugins.
# Any function in this module can add jobs to this queue but only the job thread can remove items.
_plugin_queue: Queue = Queue()

# A map from the package identifier to the UID.
# Only plugins that are currently enabled will appear in this dictionary.
_enabled_plugins: dict[str, LibraryUID] = {}

_splash_load_screen: Optional[Splash] = None
_splash_unload_screen: Optional[Splash] = None


class PluginJobThread(QThread):
    def run(self):
        while True:
            job: Optional[PluginJob] = None
            while True:
                if self.isInterruptionRequested():
                    log.debug("Exiting plugin job thread")
                    return
                try:
                    job = _plugin_queue.get(timeout=0.1)
                except Empty:
                    continue
                else:
                    break
            if job.job_type is PluginJobType.Enable:
                _enable_plugin(job.plugin_identifier)
            elif job.job_type is PluginJobType.Disable:
                _disable_plugin(job.plugin_identifier)
            elif job.job_type is PluginJobType.Reload:
                _disable_plugin(job.plugin_identifier)
                _enable_plugin(job.plugin_identifier)


class Event(QObject):
    plugin_state_change = Signal(LibraryUID, PluginState)


_event = Event()


TracePattern = re.compile(r"\s*File\s*\"(?P<path>.*?)\"")


def get_trace_paths() -> list[str]:
    return list(
        reversed(
            [
                normpath(TracePattern.match(line).group("path"))
                for line in traceback.format_stack()
            ]
        )
    )[1:]


def _validate_import(imported_name: str, frame: FrameType):
    # Plugins can only import libraries and plugins they have specified as a dependency.
    # Plugins can only be imported by other plugins.
    # We step back through the stack.
    # If we find a plugin that does not have the authority then we raise an error.
    # If we do not find a plugin in the stack we raise an error

    # Get the root module name of the module being imported. Eg a.b.c => a
    imported_root_name = imported_name.split(".")[0]

    # Find what imported the module
    while frame.f_globals.get("__name__").split(".")[0] == "importlib":
        # Skip over the import mechanisms
        frame = frame.f_back

    if frame.f_globals.get("__name__") == __name__ and frame.f_code.co_name in {
        "_enable_plugin",
        "wrap_importer_import",
    }:
        return

    importer_name = frame.f_globals.get("__name__")
    if importer_name is None:
        raise RuntimeError(f"Could not find __name__ attribute for frame\n{frame}")

    importer_root_name = importer_name.split(".")[0]
    if importer_root_name in _enabled_plugins:
        # The module was imported by a plugin
        importer_uid = _enabled_plugins[importer_root_name]
        plugin_container = _plugins[importer_uid]
        if imported_root_name in _enabled_plugins:
            # A plugin imported a plugin
            if importer_root_name != imported_root_name:
                # a plugin imported a different plugin
                if not any(
                    dependency.identifier == imported_root_name
                    for dependency in plugin_container.data.depends.plugin
                ):
                    # imported by a plugin that does not have the dependency listed
                    raise RuntimeError(
                        f"Plugin {importer_root_name} imported plugin {imported_root_name} which it does not have authority for.\nYou must list a plugin dependency in your plugin's metadata to be able to import it."
                    )
        else:
            # A plugin imported a normal module
            if (
                imported_root_name in sys.builtin_module_names
                or imported_root_name in sys.stdlib_module_names
            ):
                # Plugins don't need to specify native python libraries.
                pass
            else:
                if imported_root_name not in Packages:
                    raise RuntimeError(f"Could not find library {importer_root_name}.")
                package_name = Packages[imported_root_name][0].lower().replace("-", "_")
                if not any(
                    dependency.identifier == package_name
                    for dependency in plugin_container.data.depends.library
                ):
                    raise RuntimeError(
                        f"Plugin {importer_root_name} imported library {imported_root_name} which it does not have authority for.\nYou must list a dependency in your plugin's metadata to be able to import it."
                    )
    elif imported_root_name in _enabled_plugins:
        code = frame.f_code
        raise RuntimeError(
            f"Plugin module {imported_name} was imported by a non-plugin module {importer_name} {getattr(code, 'co_qualname', None) or getattr(code, 'co_name', 'could not resolve function name')}"
        )


class CustomSysModules(UserDict):
    def __init__(self, original: dict):
        super().__init__()
        self.data = original  # I would prefer to do this but getitem does not get called if this line is used instead.
        # self.data = copy(original)

    def __getitem__(self, imported_name: str):
        if not isinstance(imported_name, str):
            raise TypeError

        # Find the first frame before UserDict code
        frame = inspect.currentframe()
        if frame is not None:
            frame = frame.f_back
            while frame.f_globals.get("__name__") == "collections.abc":
                frame = frame.f_back

            _validate_import(imported_name, frame)

        return super().__getitem__(imported_name)

    def _remove_plugin(self, plugin_name: str):
        plugin_prefix = f"{plugin_name}."
        key: str
        for key in list(self.data.keys()):
            if key == plugin_name or key.startswith(plugin_prefix):
                del self[key]


def wrap_importer(imp):
    def wrap_importer_import(name, globals=None, locals=None, fromlist=(), level=0):
        if level:
            name_split = globals["__name__"].split(".")
            imported_name = f"{'.'.join(name_split[:len(name_split)-level+1])}.{name}"
        else:
            imported_name = name
        frame = inspect.currentframe()
        if frame is not None:
            frame = frame.f_back
            _validate_import(imported_name, frame)
        return imp(name, globals=globals, locals=locals, fromlist=fromlist, level=level)

    return wrap_importer_import


def load():
    """
    Find plugins and initialise the state.
    This must be called before any other functions in this module can be called.
    It can only be called once.
    """
    global _splash_load_screen
    log.debug("Loading plugin manager")
    log.debug("Waiting for plugin lock")
    with _plugin_lock:
        log.debug("Acquired the plugin lock")

        _splash_load_screen = Splash()
        _splash_load_screen.setModal(True)
        _splash_load_screen.show()

        # Remove the plugin directories from sys.path so that they are not directly importable
        for i in range(len(sys.path) - 1, -1, -1):
            path = sys.path[i]
            for plugin_path in plugin_dirs():
                try:
                    is_same = samefile(path, plugin_path)
                except FileNotFoundError:
                    pass
                else:
                    if is_same:
                        sys.path.pop(i)
                        break

        sys.modules = CustomSysModules(sys.modules)
        builtins.__import__ = wrap_importer(builtins.__import__)
        scan_plugins()
        plugin_state = get_plugins_state()
        for plugin_uid, plugin_container in _plugins.items():
            if plugin_state.get(plugin_uid) or plugin_container.data.locked:
                _enable_plugin(plugin_uid)
        _plugin_diagnostic()

        _splash_load_screen.close()
        _splash_load_screen = None

    log.debug("Finished loading plugins.")


def unload():
    """
    Called just before application exit to tear down all plugins.
    :return:
    """
    global _splash_unload_screen

    log.debug("Unloading plugin manager")

    log.debug("Waiting for plugin lock")
    with _plugin_lock:
        log.debug("Acquired the plugin lock")

        _splash_unload_screen = Splash()
        _splash_unload_screen.setModal(True)
        _splash_unload_screen.show()

        t = time.time()

        for plugin_uid in _plugins.keys():
            _disable_plugin(plugin_uid)

        sleep_time = 0.5 - (time.time() - t)
        if sleep_time > 0:
            time.sleep(sleep_time)
        _splash_unload_screen.close()
        _splash_unload_screen = None

    log.debug("Finished unloading plugins")


def plugin_uids() -> tuple[LibraryUID, ...]:
    """Get a tuple of all plugin unique identifiers that are installed."""
    with _plugin_lock:
        return tuple(_plugins)


def get_plugins_state() -> dict[LibraryUID, bool]:
    """
    Get the state (enabled or disabled) for each plugin.
    If a plugin is not included it defaults to False.
    """
    # TODO: load this from the global and local configuration files
    return {}


def _set_plugin_state(plugin_container: PluginContainer, plugin_state: PluginState):
    plugin_container.state = plugin_state
    uid = plugin_container.data.uid
    identifier = uid.identifier
    if plugin_state is PluginState.Enabled:
        if identifier in _enabled_plugins:
            raise RuntimeError(
                f"A plugin with identifier {identifier} has already been enabled."
            )
        _enabled_plugins[identifier] = uid
    elif identifier in _enabled_plugins:
        del _enabled_plugins[identifier]

    # self.__plugins_config[plugin_container.data.uid.to_string()] = bool(plugin_state)
    # self.__save_plugin_config()
    _event.plugin_state_change.emit(plugin_container.data.uid, plugin_container.state)


def scan_plugins():
    """
    Scan the plugin directory for newly added plugins.
    This does not load the python code. It just parses the plugin.json file and populates the plugin entry.
    """
    with _plugin_lock:
        # Find and parse all plugins
        for plugin_dir in plugin_dirs():
            for manifest_path in glob.glob(
                os.path.join(plugin_dir, "*", "plugin.json")
            ):
                try:
                    plugin_path = os.path.dirname(manifest_path)
                    plugin_container = PluginContainer.from_path(plugin_path)
                    plugin_uid = plugin_container.data.uid

                    # Ensure that the module name does not shadow an existing module
                    try:
                        mod = import_module(plugin_uid.identifier)
                    except ModuleNotFoundError:
                        # No module with this name. We are all good
                        pass
                    else:
                        # Imported a module with this name
                        if (
                            plugin_path != mod.__path__[0]
                            if hasattr(mod, "__path__")
                            else plugin_path != os.path.splitext(mod.__file__)[0]
                        ):
                            # If the path does not match the expected path then it shadows an existing module
                            log.warning(
                                f"Skipping {plugin_container.data.path} because it would shadow module {plugin_uid.identifier}."
                            )
                            continue

                    if plugin_uid not in _plugins:
                        _plugins[plugin_uid] = plugin_container
                    elif _plugins[plugin_uid].data.path != plugin_path:
                        log.warning(
                            f"Two plugins cannot have the same identifier and version.\n{_plugins[plugin_uid].data.path} and {plugin_path} have the same identifier and version."
                        )
                except Exception as e:
                    log.exception(e)


# @register_global_function
# def local_enable_plugin(plugin_uid: LibraryUID):
#     """
#     Load and initialise a plugin in the current processes.
#     Any plugins that are inactive because they depend on this plugin will also be enabled.
#     This returns immediately and is completed asynchronously.
#
#     :param plugin_uid: The plugin uid to load.
#     """
#     log.debug(f"Locally enabling plugin {plugin_uid}")
#     _plugin_queue.put(PluginJob(plugin_uid, PluginJobType.Enable))


def _has_library(requirement: Requirement):
    try:
        return version(requirement.identifier) in requirement.specifier
    except Exception:
        return False


def _has_plugin(requirement: Requirement):
    return (
        requirement.identifier in _enabled_plugins
        and _enabled_plugins[requirement.identifier] in requirement
    )


def _enable_plugin(plugin_uid: LibraryUID):
    """Enable a plugin. This must only be called by the job thread.
    :param plugin_uid: The unique identifier of the plugin to enable
    :raises: Exception if an error happened when loading the plugin.
    """
    with _plugin_lock:
        if not isinstance(plugin_uid, LibraryUID):
            raise TypeError
        plugin_container = _plugins[plugin_uid]
        if plugin_container.state is not PluginState.Disabled:
            # Cannot enable a plugin that is not currently disabled.
            return
        _set_plugin_state(plugin_container, PluginState.Inactive)

        enabled_count = -1
        while enabled_count:
            enabled_count = 0
            for plugin_container in list(_plugins.values()):
                if (
                    plugin_container.state is PluginState.Inactive
                    and PythonVersion in plugin_container.data.depends.python
                    and all(map(_has_library, plugin_container.data.depends.library))
                    and all(map(_has_plugin, plugin_container.data.depends.plugin))
                ):
                    # all dependencies are satisfied so the plugin can be enabled.
                    try:
                        _set_plugin_state(plugin_container, PluginState.Enabled)
                        log.debug(f"enabling plugin {plugin_container.data.uid}")
                        if _splash_load_screen is not None:
                            _splash_load_screen.showMessage(
                                f"Enabling plugin {plugin_container.data.uid.identifier}"
                            )
                        path = plugin_container.data.path
                        if os.path.isdir(path):
                            path = os.path.join(path, "__init__.py")

                        spec = spec_from_file_location(
                            plugin_container.data.uid.identifier, path
                        )
                        if spec is None:
                            raise Exception
                        mod = module_from_spec(spec)
                        if mod is None:
                            raise Exception
                        sys.modules[plugin_container.data.uid.identifier] = mod
                        spec.loader.exec_module(mod)

                        plugin_container.instance = mod

                        try:
                            load_plugin = plugin_container.instance.load_plugin
                        except AttributeError:
                            # The plugin does not have a load_plugin method
                            pass
                        else:
                            # User code must be run from the main thread to avoid issues.
                            invoke(load_plugin)

                        log.debug(f"enabled plugin {plugin_container.data.uid}")
                    except Exception as e:
                        log.exception(e)
                        display_exception(
                            title=f"Error while loading plugin {plugin_container.data.uid.identifier} {plugin_container.data.uid.version}",
                            error=str(e),
                            traceback=traceback.format_exc(),
                        )

                        # Since the plugin failed to load we must try and disable it
                        _disable_plugin(plugin_container.data.uid)
                    else:
                        enabled_count += 1


def _plugin_diagnostic():
    """Useful plugin diagnostic data."""
    log.debug(f"Found {len(_plugins)} plugins.")
    for plugin_container in list(_plugins.values()):
        if plugin_container.state is PluginState.Enabled:
            log.debug(f"Plugin {plugin_container.data.uid} has been enabled.")
        elif plugin_container.state is PluginState.Disabled:
            log.debug(f"Plugin {plugin_container.data.uid} is not enabled.")
        elif plugin_container.state is PluginState.Inactive:
            log.debug(
                f"Plugin {plugin_container.data.uid} could not be enabled. "
                f"PyCompatible={PythonVersion in plugin_container.data.depends.python}, "
                f"MissingLibraries={[str(l) for l in plugin_container.data.depends.library if not _has_library(l)]}, "
                f"MissingPlugins={[str(p) for p in plugin_container.data.depends.plugin if not _has_plugin(p)]}"
            )


# @register_global_function
# def local_disable_plugin(plugin_uid: LibraryUID):
#     """
#     Disable and destroy a plugin in the current process.
#     Any dependent plugins will be disabled before disabling this plugin.
#     This returns immediately and is completed asynchronously.
#
#     :param plugin_uid: The plugin uid to disable.
#     """
#     log.debug(f"Locally disabling plugin {plugin_uid}")
#     _plugin_queue.put(PluginJob(plugin_uid, PluginJobType.Disable))


def _disable_plugin(plugin_uid: LibraryUID):
    """Disable a plugin and inactive all dependents. This must only be called by the job thread."""
    with _plugin_lock:
        if not isinstance(plugin_uid, LibraryUID):
            raise TypeError
        plugin_container = _plugins[plugin_uid]
        if plugin_container.state is PluginState.Disabled:
            # Cannot disable a plugin that is already disabled
            return
        elif plugin_container.state is PluginState.Enabled:
            _unload_plugin(plugin_container)
        _set_plugin_state(plugin_container, PluginState.Disabled)


def _unload_plugin(plugin_container: PluginContainer):
    """Unload and destroy a plugin. This must only be called by the job thread."""
    _recursive_inactive_plugins(plugin_container.data.uid)
    if _splash_load_screen is not None:
        _splash_load_screen.showMessage(
            f"Disabling plugin {plugin_container.data.uid.identifier}"
        )
    elif _splash_unload_screen is not None:
        _splash_unload_screen.showMessage(
            f"Disabling plugin {plugin_container.data.uid.identifier}"
        )
    try:
        unload_plugin = plugin_container.instance.unload_plugin
    except AttributeError:
        # The plugin does not have an unload_plugin method
        pass
    else:
        # User code must be run from the main thread to avoid issues.
        try:
            invoke(unload_plugin)
        except Exception as e:
            log.exception(e)
            display_exception(
                title=f"Error while unloading plugin {plugin_container.data.uid.identifier} {plugin_container.data.uid.version}",
                error=str(e),
                traceback=traceback.format_exc(),
            )
    plugin_container.instance = None

    # Remove the module from sys.modules
    modules = sys.modules
    if isinstance(modules, CustomSysModules):
        modules._remove_plugin(plugin_container.data.uid.identifier)


def _recursive_inactive_plugins(plugin_uid: LibraryUID):
    """
    Recursively inactive all dependents of a plugin This must only be called by the job thread.
    When a plugin is disabled none of its dependents are valid any more so they must be inactivated.
    :param plugin_uid: The plugin unique identifier to find dependents of.
    """
    for plugin_container in _plugins.values():
        if plugin_container.state is PluginState.Enabled and any(
            plugin_uid in requirement
            for requirement in plugin_container.data.depends.plugin
        ):
            _unload_plugin(plugin_container)
            _set_plugin_state(plugin_container, PluginState.Inactive)


# @register_global_function
# def local_reload_plugin(plugin_uid: LibraryUID):
#     """
#     Disable a plugin if enabled and reload from source.
#     Only effects the current process.
#     This returns immediately and is completed asynchronously.
#
#     :param plugin_uid: The plugin uid to reload.
#     """
#     log.debug(f"Locally reloading plugin {plugin_uid}")
#     _plugin_queue.put(PluginJob(plugin_uid, PluginJobType.Reload))


# @register_global_function
# def global_enable_plugin(plugin_uid: LibraryUID):
#     """
#     Enable a plugin for all processes.
#     This returns immediately and is completed asynchronously.
#     """
#     log.debug(f"Globally enabling plugin {plugin_uid}")
#     if get_process_type() is ProcessType.Main:
#         # If this is the main process enable for self
#         # and tell all child processes to enable
#         local_enable_plugin(plugin_uid)
#         call_in_children(local_enable_plugin, plugin_uid)
#     elif get_process_type() is ProcessType.Child:
#         # If this is a child process then notify the main process.
#         call_in_parent(global_enable_plugin, plugin_uid)
#     else:
#         raise RuntimeError


# @register_global_function
# def global_disable_plugin(plugin_uid: LibraryUID):
#     """
#     Disable a plugin for all processes.
#     This returns immediately and is completed asynchronously.
#     """
#     log.debug(f"Globally disabling plugin {plugin_uid}")
#     if get_process_type() is ProcessType.Main:
#         # If this is the main process enable for self
#         # and tell all child processes to enable
#         local_disable_plugin(plugin_uid)
#         call_in_children(local_disable_plugin, plugin_uid)
#     elif get_process_type() is ProcessType.Child:
#         # If this is a child process then notify the main process.
#         call_in_parent(global_disable_plugin, plugin_uid)
#     else:
#         raise RuntimeError


# @register_global_function
# def global_reload_plugin(plugin_uid: LibraryUID):
#     """
#     Reload a plugin for all processes.
#     This returns immediately and is completed asynchronously.
#     """
#     log.debug(f"Globally reloading plugin {plugin_uid}")
#     if get_process_type() is ProcessType.Main:
#         # If this is the main process enable for self
#         # and tell all child processes to enable
#         local_reload_plugin(plugin_uid)
#         call_in_children(local_reload_plugin, plugin_uid)
#     elif get_process_type() is ProcessType.Child:
#         # If this is a child process then notify the main process.
#         call_in_parent(global_reload_plugin, plugin_uid)
#     else:
#         raise RuntimeError


# def install_plugin(path: str):
#     """
#     Extract a zip file containing a plugin to the dynamic plugin directory and validate its contents.
#     This will not enable or execute any of the code.
#
#     :param path: The path to a zip file containing a plugin to install.
#     :raises Exception: if the file does not meet the requirements for a plugin.
#     """
#     if get_process_type() is ProcessType.Main:
#         raise NotImplementedError
#     else:
#         raise RuntimeError("The plugin state can only be modified in the main process.")
#
#
# def uninstall_plugin(plugin_uid: LibraryUID):
#     """
#     Disable and uninstall a plugin.
#
#     :param plugin_uid: The plugin uid to uninstall.
#     """
#     if get_process_type() is ProcessType.Main:
#         raise NotImplementedError
#     else:
#         raise RuntimeError("The plugin state can only be modified in the main process.")



import os
from functools import partial
from typing import Callable, Optional, Union

import amulet
from amulet_editor.data import packages, project
from amulet_editor.models.generic import Observer
from amulet_editor.models.minecraft import LevelData
from amulet_editor.models.widgets import AIconButton
from amulet_editor.tools.programs import Programs
from amulet_editor.tools.project import Project
from amulet_editor.tools.startup._models import Menu, Navigate
from amulet_editor.tools.startup._widgets import QLevelSelectionCard
from amulet_editor.tools.startup.panels._world_selection import WorldSelectionPanel
from PySide6.QtCore import QCoreApplication, QObject, QSize
from PySide6.QtWidgets import (
    QApplication,
    QFileDialog,
    QFrame,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QVBoxLayout,
    QWidget,
)


class OpenWorldMenu(QObject):
    def __init__(self, set_panel: Callable[[Optional[QWidget]], None]):
        super().__init__()

        self.level_directory: Optional[str] = None
        self.set_panel = set_panel

        self._enable_next = Observer(bool)

        self._widget = OpenWorldWidget()
        self._widget.btn_level_directory.clicked.connect(partial(self.import_level))
        self._widget.crd_select_level.clicked.connect(partial(self.select_level))

        self._world_selection_panel = WorldSelectionPanel()
        self._world_selection_panel.level_data.connect(self.set_level)

        QApplication.instance().focusChanged.connect(self.check_focus)

    @property
    def title(self) -> str:
        return "Open World"

    @property
    def enable_next(self) -> Observer:
        return self._enable_next

    def navigated(self, destination):
        if destination == Navigate.NEXT:
            tools = packages.list_tools()
            for tool in reversed(tools):
                packages.disable_tool(tool)

            packages.enable_tool(Project())
            packages.enable_tool(Programs())
            project.set_root(self.level_directory)

    def widget(self) -> QWidget:
        return self._widget

    def next_menu(self) -> Optional[Menu]:
        return None

    def import_level(self):
        self.uncheck_level_card()

        path = self._widget.lne_level_directory.text()
        path = os.path.expanduser("~") if path == "" else path
        path = QFileDialog.getExistingDirectory(
            None,
            "Select Minecraft World",
            os.path.realpath(path),
            QFileDialog.Option.ShowDirsOnly,
        )
        self._widget.btn_level_directory.setChecked(False)
        self.set_level(path)

    def select_level(self):
        if self._widget.crd_select_level.isChecked():
            self.set_panel(self._world_selection_panel)
        else:
            self.set_panel(None)

    def set_level(self, level: Union[str, LevelData]):
        if isinstance(level, str):
            path = str(os.path.sep).join(level.split("/"))

        if isinstance(level, LevelData):
            self.level_directory = level.path

            self._widget.lne_level_directory.setText(level.path)
            self._widget.crd_select_level.setLevel(level)

            self.enable_next.emit(True)
        elif os.path.isfile(os.path.join(path, "level.dat")):
            level_data = LevelData(amulet.load_format(path))

            self.level_directory = level_data.path

            self._widget.lne_level_directory.setText(level_data.path)
            self._widget.crd_select_level.setLevel(level_data)

            self.enable_next.emit(True)
        else:
            self._widget.lne_level_directory.setText(path)
            self._widget.crd_select_level.setLevel(None)

            self.enable_next.emit(False)

    def check_focus(self, old: Optional[QWidget], new: Optional[QWidget]):
        alternate_focus = [self._widget.lne_level_directory]

        if new in alternate_focus:
            self.uncheck_level_card()

    def uncheck_level_card(self):
        self._widget.crd_select_level.setChecked(False)
        self._widget.crd_select_level.clicked.emit()


class OpenWorldWidget(QWidget):
    def __init__(self):
        super().__init__()

        self.setupUi()

    def setupUi(self):
        # Configure 'Select Level'
        self.lbl_select_level = QLabel(self)
        self.lbl_select_level.setProperty("color", "on_primary")

        self.crd_select_level = QLevelSelectionCard(self)
        self.crd_select_level.setCheckable(True)
        self.crd_select_level.setFixedHeight(105)
        self.crd_select_level.setProperty("backgroundColor", "primary")
        self.crd_select_level.setProperty("border", "surface")
        self.crd_select_level.setProperty("borderRadiusVisible", True)

        # Configure 'Level Directory'
        self.lbl_level_directory = QLabel(self)
        self.lbl_level_directory.setProperty("color", "on_primary")

        lyt_level_directory = QHBoxLayout(self)

        self.frm_level_directory = QFrame(self)
        self.frm_level_directory.setFrameShape(QFrame.Shape.NoFrame)
        self.frm_level_directory.setFrameShadow(QFrame.Shadow.Raised)
        self.frm_level_directory.setLayout(lyt_level_directory)

        self.lne_level_directory = QLineEdit(self.frm_level_directory)
        self.lne_level_directory.setFixedHeight(25)
        self.lne_level_directory.setProperty("color", "on_surface")
        self.lne_level_directory.setReadOnly(True)

        self.btn_level_directory = AIconButton("folder.svg", self.frm_level_directory)
        self.btn_level_directory.setCheckable(True)
        self.btn_level_directory.setFixedSize(QSize(27, 27))
        self.btn_level_directory.setIconSize(QSize(15, 15))
        self.btn_level_directory.setProperty("backgroundColor", "primary")

        lyt_level_directory.addWidget(self.lne_level_directory)
        lyt_level_directory.addWidget(self.btn_level_directory)
        lyt_level_directory.setContentsMargins(0, 0, 0, 0)
        lyt_level_directory.setSpacing(5)

        # Create 'Page Content' layout
        layout = QVBoxLayout(self)
        layout.addSpacing(10)
        layout.addWidget(self.lbl_select_level)
        layout.addWidget(self.crd_select_level)
        layout.addSpacing(10)
        layout.addWidget(self.lbl_level_directory)
        layout.addWidget(self.frm_level_directory)
        layout.addStretch()
        layout.setSpacing(5)

        self.setProperty("backgroundColor", "background")
        self.setLayout(layout)

        self.retranslateUi()

    def retranslateUi(self):
        # Disable formatting to condense tranlate functions
        # fmt: off
        self.lbl_select_level.setText(QCoreApplication.translate("NewProjectTypePage", "Select World", None))
        self.lbl_level_directory.setText(QCoreApplication.translate("NewProjectTypePage", "World Directory", None))
        # fmt: on



from amulet_editor import __version__
from amulet_editor.data import build
from amulet_editor.tools.startup._widgets import AIconCard
from PySide6.QtCore import QCoreApplication, QSize, Qt
from PySide6.QtGui import QImage, QPixmap
from PySide6.QtWidgets import QLabel, QVBoxLayout, QWidget


class StartupPage(QWidget):
    def __init__(self):
        super().__init__()

        self.setupUi()

        # self.frm_nav_header = QFrame(self)
        # self.frm_nav_header.setFixedHeight(25)
        # self.frm_nav_header.setFrameShape(QFrame.Shape.NoFrame)
        # self.frm_nav_header.setFrameShadow(QFrame.Shadow.Raised)
        # self.frm_nav_header.setProperty("backgroundColor", "primary")
        # self.frm_nav_header.setProperty("borderBottom", "surface")
        # self.frm_nav_header.setProperty("color", "on_surface")

    def setupUi(self):
        amulet_logo = QPixmap(QImage(build.get_resource("images/amulet_logo.png")))
        amulet_logo = amulet_logo.scaledToHeight(128)

        self.lbl_app_icon = QLabel()
        self.lbl_app_icon.setFixedHeight(128)
        self.lbl_app_icon.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.lbl_app_icon.setPixmap(amulet_logo)

        self.lbl_app_name = QLabel("Amulet Editor")
        self.lbl_app_name.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.lbl_app_name.setProperty("heading", "h1")
        self.lbl_app_name.setProperty("subfamily", "semi_light")

        self.lbl_app_version = QLabel(f"Version {__version__}")
        self.lbl_app_version.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.lbl_app_version.setProperty("color", "secondary")
        self.lbl_app_version.setProperty("heading", "h5")
        self.lbl_app_version.setProperty("subfamily", "semi_light")

        self.crd_open_level = AIconCard(
            build.get_resource(f"icons/tabler/file-symlink.svg"),
            QSize(25, 25),
            parent=self,
        )
        self.crd_open_level.setMinimumWidth(200)

        self.crd_open_project = AIconCard(
            build.get_resource(f"icons/tabler/file-symlink.svg"),
            QSize(25, 25),
            parent=self,
        )
        self.crd_open_project.setMinimumWidth(200)

        self.crd_new_project = AIconCard(
            build.get_resource(f"icons/tabler/file-plus.svg"),
            QSize(25, 25),
            parent=self,
        )
        self.crd_new_project.setMinimumWidth(200)

        layout = QVBoxLayout()
        layout.addWidget(self.lbl_app_icon)
        layout.addWidget(self.lbl_app_name)
        layout.addWidget(self.lbl_app_version)
        layout.addSpacing(20)
        layout.addWidget(self.crd_open_level)
        layout.addWidget(self.crd_open_project)
        layout.addWidget(self.crd_new_project)

        layout.setAlignment(Qt.AlignmentFlag.AlignCenter)
        layout.setAlignment(self.crd_open_level, Qt.AlignmentFlag.AlignCenter)
        layout.setAlignment(self.crd_open_project, Qt.AlignmentFlag.AlignCenter)
        layout.setAlignment(self.crd_new_project, Qt.AlignmentFlag.AlignCenter)
        layout.setSpacing(5)

        self.setLayout(layout)

        # Translate widget text
        self.retranslateUi()

    def retranslateUi(self):
        # Disable formatting to condense translate functions
        # fmt: off
        self.crd_open_level.setText(QCoreApplication.translate("StartupPage", "Open World", None))
        self.crd_new_project.setText(QCoreApplication.translate("StartupPage", "New Project", None))
        self.crd_open_project.setText(QCoreApplication.translate("StartupPage", "Open Project", None))
        # fmt: on



from __future__ import annotations

import threading
from typing import Optional
import sys
import os
import logging
from datetime import datetime
import faulthandler
from io import TextIOWrapper

from PySide6.QtCore import (
    Qt,
    QCoreApplication,
    qInstallMessageHandler,
    QtMsgType,
    QLocale,
)
from PySide6.QtWidgets import QApplication
from PySide6.QtGui import QSurfaceFormat

import amulet
import amulet_editor
from amulet_editor.models.widgets.traceback_dialog import DisplayException
from amulet_editor.data.project import _level
from amulet_editor.models.localisation import ATranslator
import amulet_editor.data._rpc as rpc

from ._cli import parse_args, BROKER
from ._app import AmuletApp
from amulet_editor.data.paths._application import _init_paths, logging_directory

log = logging.getLogger(__name__)
qt_log = logging.getLogger("Qt")


def _qt_log(msg_type, context, msg):
    if msg_type == QtMsgType.QtDebugMsg:
        qt_log.debug(msg)
    if msg_type == QtMsgType.QtInfoMsg:
        qt_log.info(msg)
    if msg_type == QtMsgType.QtWarningMsg:
        qt_log.warning(msg)
    if msg_type == QtMsgType.QtCriticalMsg:
        qt_log.critical(msg)
    if msg_type == QtMsgType.QtFatalMsg:
        qt_log.fatal(msg)


def app_main():
    args = parse_args()
    _init_paths(args.data_dir, args.config_dir, args.cache_dir, args.log_dir)

    log_file = open(
        os.path.join(
            logging_directory(),
            f"amulet-log-{datetime.now().strftime('%Y-%m-%d-%H-%M-%S')}-{os.getpid()}.txt",
        ),
        "w",
    )

    logging.basicConfig(
        level=args.logging_level,
        format=args.logging_format,
        force=True,
        handlers=[
            logging.StreamHandler(sys.__stderr__),
            logging.StreamHandler(log_file),
        ],
    )
    # TODO: remove old log files

    class StdCapture(TextIOWrapper):
        def __init__(self, logger):
            super().__init__(log_file)
            self._logger = logger

        def write(self, msg):
            if msg != "\n":
                self._logger(msg)

    # Convert all direct stdout calls (eg print) to info log calls
    sys.stdout = StdCapture(logging.getLogger("Python stdout").info)
    # Convert all direct stderr calls (eg warnings and errors) to error log calls
    sys.stderr = StdCapture(logging.getLogger("Python stderr").error)

    # Handle the qt output in a more useful way
    qInstallMessageHandler(_qt_log)

    # When running via pythonw the stderr is None so log directly to the log file
    faulthandler.enable(sys.__stderr__ or log_file)

    if args.trace:

        def trace_calls(frame, event, arg):
            if event == "call":
                try:
                    qual_name = frame.f_code.co_qualname
                    module_name = frame.f_globals["__name__"]
                    logging.info(f"Call to {module_name}.{qual_name}")
                except AttributeError:
                    pass
            return trace_calls

        sys.settrace(trace_calls)
        threading.settrace(trace_calls)

    is_broker = args.level_path == BROKER

    if is_broker:
        # Dummy application to get a main loop.
        app = QApplication()
        translator = ATranslator()
        translator.load_lang(
            QLocale(),
            "",
            directory=os.path.join(*amulet_editor.__path__, "resources", "lang"),
        )
        QCoreApplication.installTranslator(translator)
    else:
        # # Allow context sharing between widgets that do not share the same top level window.
        QCoreApplication.setAttribute(Qt.ApplicationAttribute.AA_ShareOpenGLContexts)

        # # Set the default surface format. Apparently this is required for some platforms.
        surface_format = QSurfaceFormat()
        surface_format.setDepthBufferSize(24)
        surface_format.setVersion(3, 2)
        surface_format.setProfile(QSurfaceFormat.OpenGLContextProfile.CoreProfile)
        QSurfaceFormat.setDefaultFormat(surface_format)

        app = AmuletApp()
        # The broker cannot have a level
        level_path: Optional[str] = args.level_path
        if level_path is None:
            _level.level = None
        else:
            log.debug("Loading level.")
            with DisplayException(f"Failed loading level at path {level_path}"):
                _level.level = amulet.load_level(level_path)

    rpc.init_rpc(is_broker)

    log.debug("Entering main loop.")
    exit_code = app.exec()
    log.debug(f"Exiting with code {exit_code}")
    sys.exit(exit_code)



import pathlib
from typing import Optional

from amulet_editor.data import build
from amulet_editor.models.minecraft import LevelData
from amulet_editor.models.widgets import AStylableSvgWidget
from amulet_editor.models.widgets import QElidedLabel
from PySide6.QtCore import QCoreApplication, QEvent, QSize, Qt
from PySide6.QtGui import QEnterEvent, QImage, QPixmap
from PySide6.QtWidgets import (
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QLayout,
    QPushButton,
    QSizePolicy,
    QStackedWidget,
    QWidget,
)


class AIconCard(QPushButton):
    def __init__(
        self,
        icon: str,
        icon_size: QSize,
        heading: str = "h5",
        text: str = "",
        *,
        parent: Optional[QWidget] = None,
    ):
        super().__init__(parent=parent)
        self.setProperty("hover", "false")
        layout = QHBoxLayout(self)
        layout.setContentsMargins(5, 5, 5, 5)
        layout.setAlignment(Qt.AlignmentFlag.AlignCenter)
        layout.setSpacing(5)
        self.setLayout(layout)

        self._svg_icon = AStylableSvgWidget(build.get_resource(icon))
        self._svg_icon.setAttribute(Qt.WidgetAttribute.WA_TransparentForMouseEvents)
        self._svg_icon.setSizePolicy(
            QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Maximum
        )
        self._svg_icon.setFixedSize(icon_size)
        layout.addWidget(self._svg_icon)

        self._lbl_description = QElidedLabel()
        self._lbl_description.setAttribute(
            Qt.WidgetAttribute.WA_TransparentForMouseEvents
        )
        self._lbl_description.setSizePolicy(
            QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Maximum
        )
        self._lbl_description.setProperty("subfamily", "semi_light")
        self._lbl_description.setText(text)
        self._lbl_description.setProperty("heading", heading)
        layout.addWidget(self._lbl_description)

        self.setMinimumHeight(icon_size.height() + 10)

    def setText(self, text: str):
        self._lbl_description.setText(text)

    def setHeading(self, heading: str):
        self._lbl_description.setProperty("heading", heading)

    def setIcon(self, icon_path: str):
        self._svg_icon.load(build.get_resource(icon_path))

    def enterEvent(self, event: QEnterEvent):
        self.setProperty("hover", "true")
        self.setStyleSheet("/* /")  # Force a style update.
        super().enterEvent(event)

    def leaveEvent(self, event: QEvent):
        if not self.isChecked():
            self.setProperty("hover", "false")
        self.setStyleSheet("/* /")  # Force a style update.
        super().leaveEvent(event)


class QLevelSelectionCard(QPushButton):
    def __init__(self, parent: Optional[QWidget] = None):
        if parent is None:
            super().__init__()
        else:
            super().__init__(parent=parent)

        self.setupUi()

    def setLayout(self, arg__1: QLayout):
        return None

    def setLevel(self, level_data: Optional[LevelData]):
        if level_data is not None:
            icon_path = (
                level_data.icon_path
                if level_data.icon_path is not None
                else build.get_resource("images/missing_world_icon.png")
            )
            level_icon = QPixmap(QImage(icon_path))
            level_icon = level_icon.scaledToHeight(80)

            level_name = level_data.name.get_html(font_weight=600)
            file_name = pathlib.PurePath(level_data.path).name
            version = f"{level_data.edition} - {level_data.version}"
            last_played = (
                level_data.last_played.astimezone(tz=None)
                .strftime("%B %d, %Y %I:%M %p")
                .replace(" 0", " ")
            )

            widget = QWidget(self.swgt_container)
            grid = QGridLayout(widget)

            lbl_pixmap = QLabel(widget)
            lbl_pixmap.setPixmap(level_icon)
            lbl_pixmap.setProperty("backgroundColor", "background")
            lbl_pixmap.setProperty("border", "surface")
            lbl_pixmap.setProperty("borderRadiusVisible", True)
            lbl_pixmap.setSizePolicy(
                QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Maximum
            )

            lbl_level_name = QElidedLabel(level_name, widget)
            lbl_level_name.setAttribute(Qt.WidgetAttribute.WA_TransparentForMouseEvents)

            lbl_file_name = QElidedLabel(file_name, widget)

            lbl_version = QElidedLabel(version, widget)

            lbl_last_played = QElidedLabel(last_played, widget)

            grid.addWidget(lbl_pixmap, 0, 0, 4, 1)
            grid.addWidget(lbl_level_name, 0, 1, 1, 1)
            grid.addWidget(lbl_file_name, 1, 1, 1, 1)
            grid.addWidget(lbl_version, 2, 1, 1, 1)
            grid.addWidget(lbl_last_played, 3, 1, 1, 1)

            widget.setLayout(grid)

            if self.swgt_container.widget(1) is not None:
                self.swgt_container.removeWidget(self.swgt_container.widget(1))

            self.swgt_container.addWidget(widget)
            self.swgt_container.setCurrentIndex(1)
        else:
            self.swgt_container.setCurrentIndex(0)

    def layout(self) -> QHBoxLayout:
        return super().layout()

    def setupUi(self):
        self.swgt_container = QStackedWidget(self)
        self.swgt_container.setAttribute(
            Qt.WidgetAttribute.WA_TransparentForMouseEvents
        )

        self.wgt_info = QWidget(self.swgt_container)
        self.wgt_info.setAttribute(Qt.WidgetAttribute.WA_TransparentForMouseEvents)

        self.lbl_info = QLabel(self.wgt_info)
        self.lbl_info.setProperty("color", "on_surface")
        self.lbl_info.setProperty("heading", "h5")
        self.lbl_info.setProperty("subfamily", "semi_light")

        self.lyt_info = QHBoxLayout(self.wgt_info)
        self.lyt_info.addWidget(self.lbl_info)
        self.lyt_info.setAlignment(self.lbl_info, Qt.AlignmentFlag.AlignCenter)
        self.lyt_info.setContentsMargins(0, 0, 0, 0)

        self.wgt_info.setLayout(self.lyt_info)

        self.swgt_container.addWidget(self.wgt_info)

        layout = QHBoxLayout(self)
        layout.addWidget(self.swgt_container)
        layout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(layout)

        self.retranslateUi()

    def retranslateUi(self):
        # Disable formatting to condense tranlate functions
        # fmt: off
        self.lbl_info.setText(QCoreApplication.translate("QLevelSelectionCard", "Click to Select World", None))
        # fmt: on



from amulet_editor import __version__

from PySide6.QtCore import Qt
from PySide6.QtGui import QImage, QPixmap
from amulet_editor.data import build
from ._open_world import Ui_OpenWorldPage


class OpenWorldPage(Ui_OpenWorldPage):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        back_icon = QPixmap(QImage(build.get_resource("icons/tabler/arrow-left.svg")))
        # back_icon = back_icon.scaledToHeight(30)
        self.btn_back.setIcon(back_icon)



from typing import Optional, Callable
from threading import RLock

from PySide6.QtCore import QSize, Qt
from PySide6.QtWidgets import QFrame, QWidget, QVBoxLayout, QHBoxLayout

from amulet_editor.models.widgets import ADragContainer

from ._icon import ATooltipIconButton


class AToolBar(QFrame):
    """
    A toolbar is a strip of buttons.
    The first half can be rearranged and the second half are fixed.
    """

    def __init__(
        self,
        parent: Optional[QWidget] = None,
        f: Qt.WindowType = Qt.WindowType.Widget,
        orientation=Qt.Orientation.Vertical,
    ):
        super().__init__(parent, f)

        layout_cls = {
            Qt.Orientation.Vertical: QVBoxLayout,
            Qt.Orientation.Horizontal: QHBoxLayout,
        }[orientation]

        self._lyt_main = layout_cls()
        self._lyt_main.setSpacing(5)
        self._lyt_main.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self._lyt_main)

        self._wgt_dynamic_tools = ADragContainer(self, orientation)
        self._wgt_dynamic_tools.setObjectName("_wgt_dynamic_tools")
        self._lyt_main.addWidget(self._wgt_dynamic_tools)

        self._lyt_fixed_tools = layout_cls()
        self._lyt_fixed_tools.setObjectName("_lyt_fixed_tools")
        self._lyt_fixed_tools.addStretch()
        self._lyt_main.addLayout(self._lyt_fixed_tools)

        self._buttons: dict[str, ATooltipIconButton] = {}
        self._lock = RLock()

    def _make_button(
        self, uid: str, icon: str, name: str, callback: Callable[[], None] = None
    ) -> ATooltipIconButton:
        if uid in self._buttons:
            raise ValueError(
                f"A button with unique identifier {uid} has already been registered"
            )

        button = ATooltipIconButton(icon, self)
        button.setToolTip(name)
        button.setFixedSize(QSize(40, 40))
        button.setIconSize(QSize(30, 30))

        def on_click(evt):
            if callback is not None:
                try:
                    callback()
                except Exception:
                    pass

        button.clicked.connect(on_click)
        self._buttons[uid] = button
        return button

    def add_static_button(
        self, uid: str, icon: str, name: str, callback: Callable[[], None] = None
    ):
        """
        Add a static button.
        This is reserved for internal use only.

        :param uid: A unique identifier. Used to remove this button.
        :param icon: The icon the button should use.
        :param name: The name of the button in the tool tip.
        :param callback: The function to call when the button is clicked.
        """
        with self._lock:
            button = self._make_button(uid, icon, name, callback)
            self._lyt_fixed_tools.insertWidget(1, button)

    def add_dynamic_button(
        self, uid: str, icon: str, name: str, callback: Callable[[], None] = None
    ):
        """
        Add a dynamic button.
        Plugins can add their own dynamic buttons.

        :param uid: A unique identifier. Used to remove this button.
        :param icon: The icon the button should use.
        :param name: The name of the button in the tool tip.
        :param callback: The function to call when the button is clicked.
        """
        with self._lock:
            button = self._make_button(uid, icon, name, callback)
            self._wgt_dynamic_tools.add_item(button)

    def remove_button(self, uid: str):
        """
        Remove a button from the toolbar

        :param uid: The uid of the button to remove.
        """
        with self._lock:
            button = self._buttons.pop(uid)
            button.deleteLater()

    def activate(self, uid: str):
        """
        Visually show that a tool is active.
        All other tools will be visually disabled.
        This is only a visual update and does not produce any events.

        :param uid: The uid of the button to activate.
        """
        with self._lock:
            new_button = self._buttons[uid]
            for button in self._buttons.values():
                if button is not new_button:
                    button.setChecked(False)
                    button.setCheckable(False)
            new_button.setCheckable(True)
            new_button.setChecked(True)



import os


def locate_file(file: str, directory: str, recursive=False) -> list[str]:
    """
    Returns a list of all paths containing the specified file within
    the specified directory.
    """

    paths: list[str] = []
    for item in os.listdir(directory):
        if os.path.isfile(os.path.join(directory, item)) and item == file:
            paths.append(directory + item)
        elif recursive and os.path.isdir(directory + item):
            locate_file(file, directory + item, paths)

    return paths


def locate_levels(directories: list[str]) -> list[str]:
    """
    Returns a list of paths to all identified minecraft level.dat
    files in the list of provided directories.
    """

    worlds: list[str] = []

    for directory in directories:
        for folder in os.listdir(directory):
            folder = os.path.join(directory, folder)
            if os.path.exists(os.path.join(folder, "level.dat")):
                worlds.append(os.path.join(folder))

    return worlds



from threading import Lock
from weakref import WeakKeyDictionary
from typing import Generic, TypeVar


T = TypeVar("T")


class KeyLock(Generic[T]):
    """
    A map from weakly referenced keys to locks per key.
    This is useful if you want to lock access to individual resources.
    """

    def __init__(self):
        self._lock = Lock()
        self._locks: WeakKeyDictionary[T, Lock] = WeakKeyDictionary()

    def get(self, key: T) -> Lock:
        with self._lock:
            lock = self._locks.get(key)
            if lock is None:
                lock = self._locks[key] = Lock()
            return lock



from PySide6.QtCore import Slot, QSize, Qt
from PySide6.QtGui import QGuiApplication, QIcon

from ._traceback_dialog import Ui_AmuletTracebackDialog
from amulet_editor.data.build import get_resource


class _AmuletTracebackDialog(Ui_AmuletTracebackDialog):
    """A dialog to display tracebacks."""

    def __init__(
        self,
        parent=None,
        f=Qt.WindowType.Widget,
        title: str = "",
        error: str = "",
        traceback: str = "",
    ):
        super().__init__(parent, f)
        self._traceback = traceback
        alert_icon = QIcon(get_resource("icons/tabler/alert-circle.svg")).pixmap(
            QSize(32, 32)
        )
        self._alert_image.setPixmap(alert_icon)
        self._copy_button.clicked.connect(self._on_copy)
        self.setWindowTitle(title)
        self._error_text.setText(error)
        self._traceback_text.setText(traceback)

    @Slot()
    def _on_copy(self):
        clipboard = QGuiApplication.clipboard()
        clipboard.setText(self._traceback)



from typing import Type, Callable

from amulet_team_main_window import _plugin
from .models.view import View
from .application.windows.main_window import AmuletMainWindow


UID = str


def get_windows() -> tuple[AmuletMainWindow, ...]:
    return (_plugin.window,)


def get_active_window() -> AmuletMainWindow:
    return _plugin.window


def register_view(view_cls: Type[View], icon: str, name: str):
    for window in get_windows():
        window.register_view(view_cls, icon, name)


def unregister_view(view_cls: Type[View]):
    for window in get_windows():
        window.unregister_view(view_cls)


def add_toolbar_button(
    uid: UID, icon: str, name: str, callback: Callable[[], None] = None
):
    """
    Add an icon to the toolbar for all windows.

    :param uid: The unique identifier of the button. Eg: author_name:button_name
    :param icon: The icon path to use in the toolbar.
    :param name: The name of the view to use in the icon tooltip.
    :param callback: The function to call when the button is clicked.
    :return:
    """
    for window in get_windows():
        window.add_toolbar_button(uid, icon, name, callback)


def add_static_toolbar_button(
    uid: UID, icon: str, name: str, callback: Callable[[], None] = None
):
    """
    Add a static icon to the toolbar for all windows.
    These should be reserved for special cases.
    Third party plugins should use :func:`add_button`.

    :param uid: The unique identifier of the button. Eg: author_name:button_name
    :param icon: The icon path to use in the toolbar.
    :param name: The name of the view to use in the icon tooltip.
    :param callback: The function to call when the button is clicked.
    :return:
    """
    for window in get_windows():
        window.add_static_toolbar_button(uid, icon, name, callback)


def remove_toolbar_button(uid: UID):
    """
    Remove a toolbar button from all windows.

    :param uid: The unique identifier for the toolbar button to remove.
    :return:
    """
    for window in get_windows():
        window.remove_toolbar_button(uid)



import os
from typing import Optional

from PySide6.QtCore import QStandardPaths


DefaultDataDir = os.path.join(
    QStandardPaths.writableLocation(QStandardPaths.StandardLocation.AppDataLocation),
    "AmuletTeam",
    "AmuletEditor",
    "data",
)
DefaultConfigDir = os.path.join(
    QStandardPaths.writableLocation(QStandardPaths.StandardLocation.AppDataLocation),
    "AmuletTeam",
    "AmuletEditor",
    "config",
)
DefaultCacheDir = os.path.join(
    QStandardPaths.writableLocation(QStandardPaths.StandardLocation.AppDataLocation),
    "AmuletTeam",
    "AmuletEditor",
    "cache",
)
DefaultLogDir = os.path.join(
    QStandardPaths.writableLocation(QStandardPaths.StandardLocation.AppDataLocation),
    "AmuletTeam",
    "AmuletEditor",
    "logs",
)


def _init_paths(
    data_dir: Optional[str],
    config_dir: Optional[str],
    cache_dir: Optional[str],
    log_dir: Optional[str],
):
    if data_dir is None:
        os.environ.setdefault("DATA_DIR", DefaultDataDir)
    else:
        os.environ["DATA_DIR"] = data_dir

    if config_dir is None:
        os.environ.setdefault("CONFIG_DIR", DefaultConfigDir)
    else:
        os.environ["CONFIG_DIR"] = config_dir

    if cache_dir is None:
        os.environ.setdefault("CACHE_DIR", DefaultCacheDir)
    else:
        os.environ["CACHE_DIR"] = cache_dir

    if log_dir is None:
        os.environ.setdefault("LOG_DIR", DefaultLogDir)
    else:
        os.environ["LOG_DIR"] = log_dir


def data_directory() -> str:
    """
    Returns a path to the directory used for storage of persistent data.
    Generates appropriate directories if path does not already exist.
    """
    directory = os.environ["DATA_DIR"]
    os.makedirs(directory, exist_ok=True)
    return directory


def config_directory() -> str:
    """
    Returns a path to the directory used for storage of configuration data.
    Generates appropriate directories if path does not already exist.
    """
    directory = os.environ["CONFIG_DIR"]
    os.makedirs(directory, exist_ok=True)
    return directory


def cache_directory() -> str:
    """
    Returns a path to the directory used for storage of cache data.
    Generates appropriate directories if path does not already exist.
    """
    directory = os.environ["CACHE_DIR"]
    os.makedirs(directory, exist_ok=True)
    return directory


def logging_directory() -> str:
    """
    Returns a path to the directory used for storage of logging data.
    Generates appropriate directories if path does not already exist.
    """
    directory = os.environ["LOG_DIR"]
    os.makedirs(directory, exist_ok=True)
    return directory


def user_directory() -> str:
    directory = os.path.join(data_directory(), "user")
    os.makedirs(directory, exist_ok=True)

    return directory


def project_directory(project_name: Optional[str] = None) -> str:
    """Returns a path to the default location for storing Amulet projects."""

    documents = os.path.normpath(
        QStandardPaths.writableLocation(
            QStandardPaths.StandardLocation.DocumentsLocation
        )
    )

    directory = os.path.join(
        documents,
        "Amulet",
        "projects",
    )

    if project_name is not None:
        directory = os.path.join(directory, project_name)

    os.makedirs(directory, exist_ok=True)
    return directory



from __future__ import annotations

from typing import Generic, TypeVar, Callable, Any, Generator, Union
from threading import Lock
from enum import IntEnum
from PySide6.QtCore import QObject, Signal, QThreadPool, SignalInstance
from amulet_editor.data.dev._debug import enable_trace


T = TypeVar("T")
Void = object()


class Promise(QObject, Generic[T]):
    """
    A class to simplify asynchronous function calls.
    A target function is passed in which is run in a new thread.
    The target function is given the promise instance through which it can notify progress changes and text changes.
    The return value from the function is accessible from :meth:`get_return` or if an exception occurred it will be re-raised when :meth:`get_return` is called.
    It can also optionally support canceling by periodically checking is_cancel_requested and raising :attr:`OperationCanceled` to signify its acceptance of the cancel.
    """

    class Data:
        def __init__(
            self,
            progress_change: SignalInstance,
            progress_text_change: SignalInstance,
            is_cancel_requested: Callable[[], bool],
        ):
            self.progress_change = progress_change
            self.progress_text_change = progress_text_change
            self.is_cancel_requested = is_cancel_requested

    class Status(IntEnum):
        NotStarted = 0
        Running = 1
        Finished = 2
        Canceled = 3

    class OperationCanceled(Exception):
        pass

    def __init__(self, target: Callable[[Promise.Data], T]):
        """
        Create a new promise
        :param target: The function to execute.
        """
        super().__init__()
        self._lock = Lock()
        self._status = self.Status.NotStarted
        self._cancel_requested = False
        self._target = target
        self._value: T = Void
        self._exception = Void

    ready = Signal()
    canceled = Signal()
    progress_change = Signal(float)
    progress_text_change = Signal(str)

    def _op_wrapper(self):
        try:
            value = self._target(
                Promise.Data(
                    self.progress_change,
                    self.progress_text_change,
                    self.is_cancel_requested,
                )
            )
        except self.OperationCanceled as e:
            self._exception = e
            self._status = self.Status.Canceled
            self.canceled.emit()
        except Exception as e:
            self._exception = e
            self._status = self.Status.Finished
            self.ready.emit()
        else:
            self._value = value
            self._status = self.Status.Finished
            self.ready.emit()

    def _thread_op_wrapper(self):
        enable_trace()
        self._op_wrapper()

    def _set_start(self):
        with self._lock:
            if self._status is not self.Status.NotStarted:
                raise RuntimeError("This promise has already been started.")
            self._status = self.Status.Running

    def start(self):
        """
        Run the operation wrapped by the promise in a new thread.
        This will return immediately.
        This can only be called once.
        """
        self._set_start()
        QThreadPool.globalInstance().start(self._thread_op_wrapper)

    def call(self) -> T:
        """
        Directly call the operation wrapped by the promise in this thread.
        Block until complete.
        Exceptions will be raised to the caller.
        This can only be called once.
        """
        self._set_start()
        self._op_wrapper()
        return self.get_return()

    def call_chained(
        self,
        parent_promise: Promise.Data,
        min_progress: float = 0.0,
        max_progress: float = 1.0,
    ) -> T:
        """
        Directly call the operation wrapped by this promise in this thread.
        Relay signals to another promise.
        Block until complete.
        This can only be called once.

        :param parent_promise: The parent promise to relay the signals to.
        :param min_progress: The minimum progress when relaying progress to the parent promise.
        :param max_progress: The maximum progress when relaying progress to the parent promise.
        :return:
        """
        self._set_start()

        progress_multiplier = max_progress - min_progress

        def tick(progress: float):
            parent_promise.progress_change.emit(
                progress * progress_multiplier + min_progress
            )
            if parent_promise.is_cancel_requested():
                self.cancel()

        self.progress_change.connect(tick)
        self.progress_text_change.connect(parent_promise.progress_text_change)

        self._op_wrapper()
        return self.get_return()

    def cancel(self):
        """
        Request the operation be canceled.
        It is down to the operation to implement support for this.
        """
        self._cancel_requested = True

    def is_cancel_requested(self) -> bool:
        """Has cancel been called to signal that the operation should be canceled."""
        return self._cancel_requested

    def status(self) -> Status:
        return self._status

    def get_return(self) -> T:
        """Get the return value or raise an exception if one occurred."""
        if self._value is not Void:
            return self._value
        elif self._exception is not Void:
            raise self._exception
        else:
            raise RuntimeError("The operation has not finished yet")



from __future__ import annotations
from typing import Optional, Union as Intersection, Union
from enum import IntEnum

from PySide6.QtWidgets import (
    QWidget,
    QApplication,
    QStackedWidget,
    QVBoxLayout,
    QHBoxLayout,
    QPushButton,
    QScrollArea,
    QSizePolicy,
    QFrame,
    QLabel,
    QSplitter,
)
from PySide6.QtGui import (
    QIcon,
    QPixmap,
    QPainter,
    QMouseEvent,
    QResizeEvent,
    QWheelEvent,
    QColor,
    QPaintEvent,
    QPolygon,
    QCursor,
)
from PySide6.QtCore import (
    Signal,
    Slot,
    QPoint,
    Qt,
    QObject,
    QSize,
    QEvent,
    QRect,
)

from amulet_editor.data.build import get_resource
from amulet_editor.models.widgets.traceback_dialog import DisplayException

import amulet_team_main_window2.application.windows.sub_window as sub_window
import amulet_team_main_window2.application.windows.main_window as main_window

_button_size: Optional[QSize] = None


def button_size() -> QSize:
    global _button_size
    if _button_size is None:
        _button_size = QPushButton().sizeHint()
    return _button_size


class TabPage:
    @property
    def name(self) -> str:
        return "TabPageNameGoesHere"

    @property
    def icon(self) -> Optional[QIcon]:
        return None


class TabEngineTabButton(QFrame):
    """A tab button."""

    def __init__(self, label: str, icon: Optional[QIcon]):
        super().__init__()
        self.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Fixed)
        self.layout = QHBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.layout)
        self.layout.addStretch()
        self.layout.addWidget(QLabel(label))
        self.layout.addStretch()
        self.close_button = QPushButton()
        size = int(button_size().height() * 0.75)
        self.close_button.setFixedSize(size, size)
        self.close_button.setIcon(QIcon(get_resource("icons/tabler/x.svg")))
        self.close_button.setFlat(True)
        self.layout.addWidget(self.close_button)

    def display_clicked(self):
        self.setFrameStyle(QFrame.Shape.Panel | QFrame.Shadow.Sunken)

    def display_unclicked(self):
        self.setFrameStyle(QFrame.Shape.Panel | QFrame.Shadow.Raised)

    def sizeHint(self) -> QSize:
        return QSize(-1, button_size().height())


class DropArea(IntEnum):
    Top = 0
    Right = 1
    Bottom = 2
    Left = 3
    Middle = 4


class DragSplitRenderer(QWidget):
    """A class to implement custom drawing while the user is dragging a tab."""

    def __init__(self, target: QWidget):
        super().__init__(target)
        self.target = target
        self.setWindowFlags(Qt.WindowType.FramelessWindowHint)
        self.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)
        self.setAttribute(Qt.WidgetAttribute.WA_TransparentForMouseEvents)

        self.shape = QSize(-1, -1)
        self.polygons: list[QPolygon] = []
        self.drop_area: Optional[DropArea] = None

        self.move(0, 0)
        self.resize(target.size())
        self._compute_polygons()
        self.show()

    def _compute_polygons(self):
        shape = self.size()
        if shape != self.shape:
            width = self.width()
            height = self.height()
            inset_amount = 3.5
            inset_width = int(width / inset_amount)
            inset_height = int(height / inset_amount)

            top_left = QPoint(0, 0)
            top_right = QPoint(width, 0)
            bottom_right = QPoint(width, height)
            bottom_left = QPoint(0, height)

            middle_top_left = QPoint(inset_width, inset_height)
            middle_top_right = QPoint(width - inset_width, inset_height)
            middle_bottom_right = QPoint(width - inset_width, height - inset_height)
            middle_bottom_left = QPoint(inset_width, height - inset_height)

            self.polygons = [
                QPolygon([top_left, top_right, middle_top_right, middle_top_left]),
                QPolygon(
                    [top_right, bottom_right, middle_bottom_right, middle_top_right]
                ),
                QPolygon(
                    [bottom_right, bottom_left, middle_bottom_left, middle_bottom_right]
                ),
                QPolygon([bottom_left, top_left, middle_top_left, middle_bottom_left]),
                QPolygon(
                    [
                        middle_top_left,
                        middle_top_right,
                        middle_bottom_right,
                        middle_bottom_left,
                    ]
                ),
            ]

            self.shape = shape

    def resizeEvent(self, event: QResizeEvent):
        self._compute_polygons()

    def paintEvent(self, event: QPaintEvent):
        painter = QPainter(self)

        cursor_point = QCursor.pos() - self.mapToGlobal(QPoint(0, 0))

        normal_colour = QColor(115, 215, 255, 128)
        painter.setBrush(normal_colour)
        painter.setPen(QColor(115, 215, 255))
        self.drop_area = None

        for drop, poly in zip(DropArea, self.polygons):
            if self.drop_area is None and poly.containsPoint(
                cursor_point, Qt.FillRule.OddEvenFill
            ):
                painter.setBrush(QColor(115, 215, 255, 190))
                painter.drawPolygon(poly)
                painter.setBrush(normal_colour)
                self.drop_area = drop
            else:
                painter.drawPolygon(poly)
            painter.drawPolyline(poly)

        painter.end()


class DragTabRenderer(QWidget):
    """A class to implement custom drawing while the user is dragging a tab."""

    def __init__(self, target: QWidget):
        super().__init__(target)
        self.target = target
        self.setWindowFlags(Qt.WindowType.FramelessWindowHint)
        self.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)
        self.setAttribute(Qt.WidgetAttribute.WA_TransparentForMouseEvents)

        self.drop_index = 0

        self.move(0, 0)
        self.resize(target.size())
        self.show()

    def paintEvent(self, event: QPaintEvent):
        painter = QPainter(self)

        normal_colour = QColor(115, 215, 255, 128)
        painter.setBrush(normal_colour)
        painter.drawRect(self.rect())
        painter.end()


class TabEngineTabContainerWidget(QWidget):
    """
    A widget containing tab buttons.
    This represents the whole area including the area off screen.
    """

    tab_changed = Signal(int)

    drag_start_pos: QPoint
    active_button: Optional[TabEngineTabButton]

    # The widget being dragged or None
    dragged_widget: Optional[Intersection[QWidget, TabPage]]
    # The widget that was previously highlighted
    highlight_widget: Union[
        None,
        tuple[TabEngineStackedTabWidget, DragSplitRenderer],
        tuple[TabEngineTabContainerWidget, DragTabRenderer],
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setAcceptDrops(True)
        self.drag_start_pos = QPoint()
        self.layout = QHBoxLayout(self)
        self.layout.setSpacing(0)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.active_button = None
        self.dragged_widget = None
        self.highlight_widget = None

    def add_tab(self, label: str, icon: QIcon = None):
        tab = TabEngineTabButton(label, icon)
        if self.active_button is None:
            self.active_button = tab
            tab.display_clicked()
        else:
            tab.display_unclicked()
        self.layout.addWidget(tab)

    def insert_tab(self, index: int, label: str, icon: QIcon = None):
        tab = TabEngineTabButton(label, icon)
        if self.active_button is None:
            self.active_button = tab
            tab.display_clicked()
        else:
            tab.display_unclicked()
        self.layout.insertWidget(index, tab)

    def remove_tab(self, index: int):
        item = self.layout.itemAt(index)
        widget = item.widget()
        self.layout.removeItem(item)
        is_active = widget == self.active_button
        widget.hide()
        widget.deleteLater()
        if is_active:
            if self.count():
                active_index = min(self.count() - 1, index)
                item = self.layout.itemAt(active_index)
                self.active_button = item.widget()
                self.active_button.display_clicked()
            else:
                self.active_button = None

    def count(self) -> int:
        return self.layout.count()

    def current_index(self) -> int:
        return self.layout.indexOf(self.active_button)

    @property
    def container(self) -> TabEngineTabContainer:
        parent = self.parent().parent()
        if not isinstance(parent, TabEngineTabContainer):
            raise RuntimeError(
                "Parent of TabEngineTabContainerWidget must be TabEngineTabContainer"
            )
        return parent

    def _get_button_at(self, point: QPoint) -> Optional[TabEngineTabButton]:
        child = self.childAt(point)
        if child is None:
            return
        while child.parent() != self:
            child = child.parent()
        if isinstance(child, TabEngineTabButton):
            return child

    def mousePressEvent(self, event: QMouseEvent):
        # If left-clicked
        if event.button() == Qt.MouseButton.LeftButton:
            # Store the click position
            self.drag_start_pos = event.position().toPoint()
            # Get the tab button that was clicked (if any)
            button = self._get_button_at(self.drag_start_pos)
            if button is None:
                return
            # Deactivate the previously clicked button
            old_button = self.active_button
            if old_button is not None:
                old_button.display_unclicked()
            # Active the clicked button
            button.display_clicked()
            self.active_button = button
            # Notify of tab change
            self.tab_changed.emit(self.layout.indexOf(button))
        else:
            super().mousePressEvent(event)

    def mouseMoveEvent(self, event: QMouseEvent):
        # If the mouse has been clicked and dragged beyond the minimum distance then start dragging
        if self.dragged_widget is not None:
            # Drag has already started. Update the widget display.
            widget = self._get_drop_widget(event.globalPosition().toPoint())
            if self.highlight_widget is not None:
                old_widget, render_widget = self.highlight_widget
                if old_widget is widget:
                    # Draw hook already installed. Just update it
                    render_widget.update()
                else:
                    # Not hovering over this widget anymore
                    render_widget.close()
                    self.highlight_widget = None

            if self.highlight_widget is None:
                if isinstance(widget, TabEngineStackedTabWidget):
                    render_widget = DragSplitRenderer(widget)
                    self.highlight_widget = widget, render_widget
                elif isinstance(widget, TabEngineTabContainerWidget):
                    render_widget = DragTabRenderer(widget)
                    self.highlight_widget = widget, render_widget

        elif (
            event.buttons() & Qt.MouseButton.LeftButton
            and self.active_button is not None
            and not self.drag_start_pos.isNull()
            and (
                (event.position() - self.drag_start_pos).manhattanLength()
                >= QApplication.startDragDistance()
            )
        ):
            # The user has clicked and dragged beyond the minimum distance
            # Get the tab display
            pixmap = self.active_button.grab()
            target_pixmap = QPixmap(pixmap.size())
            target_pixmap.fill(Qt.GlobalColor.transparent)
            painter = QPainter(target_pixmap)
            painter.setOpacity(0.85)
            painter.drawPixmap(0, 0, pixmap)
            painter.end()

            # Remove the tab
            tab_index = self.layout.indexOf(self.active_button)
            self.dragged_widget = self.container.tab_bar.tab_widget.remove_page(
                tab_index
            )
            # Start drag. This widget will get all mouse events.
            self.grabMouse(target_pixmap)
        else:
            super().mouseMoveEvent(event)

    def _get_drop_widget(
        self, point: QPoint
    ) -> Union[None, TabEngineTabContainerWidget, TabEngineStackedTabWidget]:
        """Get the widget that the dragged widget will be dropped into."""
        widget = QApplication.widgetAt(point)
        while widget is not None:
            if isinstance(
                widget, (TabEngineTabContainerWidget, TabEngineStackedTabWidget)
            ):
                return widget
            widget = widget.parent()
        return None

    def mouseReleaseEvent(self, event: QMouseEvent):
        if self.dragged_widget is None:
            # If not dragging. Reset values
            self.drag_start_pos = QPoint()
        elif not event.buttons() & Qt.MouseButton.LeftButton:
            # If dragging and left mouse has been released. Drop the widget
            # Find where the drop happened and update the widget
            self.releaseMouse()

            widget, render_widget = self.highlight_widget or (None, None)
            if isinstance(widget, TabEngineTabContainerWidget) and isinstance(
                render_widget, DragTabRenderer
            ):
                # Dropped into a tab bar
                widget.container.tab_bar.tab_widget.add_page(self.dragged_widget)
            elif isinstance(widget, TabEngineStackedTabWidget) and isinstance(
                render_widget, DragSplitRenderer
            ):
                drop_area = render_widget.drop_area
                if drop_area is DropArea.Middle:
                    widget.add_page(self.dragged_widget)
                else:
                    # Dropped onto a stacked tab widget. Replace the widget with a splitter
                    splitter = widget.splitter
                    splitter_index = splitter.indexOf(widget)
                    sizes = splitter.sizes()

                    widget.setParent(None)

                    new_splitter = RecursiveSplitter()

                    new_splitter.addWidget(widget)

                    tab_widget = TabEngineStackedTabWidget()
                    tab_widget.add_page(self.dragged_widget)
                    if drop_area in {DropArea.Top, DropArea.Bottom}:
                        new_splitter.setOrientation(Qt.Orientation.Vertical)
                    else:
                        new_splitter.setOrientation(Qt.Orientation.Horizontal)
                    if drop_area in {DropArea.Top, DropArea.Left}:
                        new_splitter.insertWidget(0, tab_widget)
                    else:
                        new_splitter.addWidget(tab_widget)

                    splitter.insertWidget(splitter_index, new_splitter)
                    splitter.setSizes(sizes)
                    new_splitter.setSizes([2048, 2048])
            else:
                # Dropped into space or a non-compatible widget
                new_window = sub_window.AmuletSubWindow(
                    main_window.AmuletMainWindow.main_window()
                )
                tab_widget = new_window.splitter_widget.widget(0)
                tab_widget.add_page(self.dragged_widget)
                new_window.move(event.globalPosition().toPoint())
                new_window.show()

            if render_widget is not None:
                render_widget.close()

            self.container.tab_bar.tab_widget.clean_up()
            self.drag_start_pos = QPoint()
            self.dragged_widget = None
            self.highlight_widget = None
        else:
            super().mouseReleaseEvent(event)


class TabEngineTabContainer(QScrollArea):
    """A scrollable widget to contain tab buttons."""

    tab_changed = Signal(int)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.widget = TabEngineTabContainerWidget()
        self.widget.tab_changed.connect(self.tab_changed)
        self.setWidget(self.widget)

        self.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        self.setWidgetResizable(True)
        self.setFrameShape(QFrame.Shape.NoFrame)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.verticalScrollBar().hide()
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.horizontalScrollBar().hide()

    @property
    def tab_bar(self) -> TabEngineTabBar:
        parent = self.parent()
        if not isinstance(parent, TabEngineTabBar):
            raise RuntimeError(
                "Parent of TabEngineTabContainer must be TabEngineTabBar"
            )
        return parent

    def sizeHint(self) -> QSize:
        return self.widget.layout.sizeHint()

    def wheelEvent(self, event: QWheelEvent):
        scroll_bar = self.horizontalScrollBar()
        scroll_bar.setValue(scroll_bar.value() - event.angleDelta().y())


class TabEngineTabBar(QWidget):
    """A custom class that behaves like a QTabBar."""

    tab_changed = Signal(int)
    add_clicked = Signal()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.layout = QHBoxLayout(self)
        self.layout.setSpacing(0)
        self.layout.setContentsMargins(0, 0, 0, 0)

        button_height = button_size().height()

        self.left_button = QPushButton("<")
        self.left_button.setFixedSize(button_height, button_height)
        self.layout.addWidget(self.left_button)
        self.left_button.clicked.connect(self._move_left)

        self.tab_container = TabEngineTabContainer()
        self.tab_container.tab_changed.connect(self.tab_changed)
        self.layout.addWidget(self.tab_container)

        self.right_button = QPushButton(">")
        self.right_button.setFixedSize(button_height, button_height)
        self.layout.addWidget(self.right_button)
        self.right_button.clicked.connect(self._move_right)

        self.plus_button = QPushButton("+")
        self.plus_button.setFixedSize(button_height, button_height)
        self.layout.addWidget(self.plus_button)
        self.plus_button.clicked.connect(self.add_clicked)

    def add_tab(self, label: str, icon: QIcon = None):
        self.tab_container.widget.add_tab(label, icon)
        self._check_size()

    def insert_tab(self, index: int, label: str, icon: QIcon = None):
        self.tab_container.widget.insert_tab(index, label, icon)
        self._check_size()

    def remove_tab(self, index: int):
        self.tab_container.widget.remove_tab(index)
        self._check_size()

    def count(self) -> int:
        return self.tab_container.widget.count()

    def current_index(self) -> int:
        return self.tab_container.widget.current_index()

    @property
    def tab_widget(self) -> TabEngineStackedTabWidget:
        parent = self.parent()
        if not isinstance(parent, TabEngineStackedTabWidget):
            raise RuntimeError("Parent of TabEngineTabBar must be TabEngineTabWidget")
        return parent

    def _check_size(self):
        if (
            self.tab_container.size().width()
            + self.left_button.width() * self.left_button.isVisible()
            + self.right_button.width() * self.right_button.isVisible()
            > self.tab_container.sizeHint().width()
        ):
            self.left_button.hide()
            self.right_button.hide()
        else:
            self.left_button.show()
            self.right_button.show()

    @Slot()
    def _move_left(self):
        scroll_bar = self.tab_container.horizontalScrollBar()
        scroll_bar.setValue(scroll_bar.value() - button_size().width() * 2)

    @Slot()
    def _move_right(self):
        scroll_bar = self.tab_container.horizontalScrollBar()
        scroll_bar.setValue(scroll_bar.value() + button_size().width() * 2)

    def resizeEvent(self, event: QResizeEvent):
        super().resizeEvent(event)
        self._check_size()


class TabEngineStackedTabWidget(QWidget):
    """A custom class that behaves like a QTabWidget"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setAcceptDrops(True)
        self._layout = QVBoxLayout()
        self.setLayout(self._layout)

        self.tab_bar = TabEngineTabBar()
        self._layout.addWidget(self.tab_bar)
        self.stacked_widget = QStackedWidget()
        self._layout.addWidget(self.stacked_widget)

        self.tab_bar.tab_changed.connect(self.stacked_widget.setCurrentIndex)
        self.tab_bar.add_clicked.connect(self._add)

    @Slot()
    def _add(self):
        # from PySide6.QtWidgets import QLabel
        # i = random.randint(0, 9)
        # self.add_tab(QLabel(f"Test Widget {i}"), "test" + "a"*random.randint(0, 20) + str(i))
        pass

    def _validate_page(
        self, page: Intersection[QWidget, TabPage]
    ) -> tuple[str, Optional[QIcon]]:
        with DisplayException(f"Error in {page.__class__}"):
            if not (isinstance(page, QWidget) and isinstance(page, TabPage)):
                raise TypeError(
                    f"Page of type {page.__class__} must be an instace of QWidget and TabPage"
                )
            name = page.name
            icon = page.icon
            if not isinstance(name, str):
                raise TypeError("Page name is invalid")
            if not (icon is None or isinstance(icon, QIcon)):
                raise TypeError("Page icon is invalid")
        return name, icon

    def add_page(self, page: Intersection[QWidget, TabPage]):
        name, icon = self._validate_page(page)
        self.tab_bar.add_tab(name, icon)
        self.stacked_widget.addWidget(page)

    def insert_page(self, index: int, page: Intersection[QWidget, TabPage]):
        name, icon = self._validate_page(page)
        self.tab_bar.insert_tab(index, name, icon)
        self.stacked_widget.insertWidget(index, page)

    def remove_page(self, index: int) -> Intersection[QWidget, TabPage]:
        self.tab_bar.remove_tab(index)
        widget = self.stacked_widget.widget(index)
        self.stacked_widget.removeWidget(widget)
        widget.setParent(None)
        # clean_up_widgets(self)
        return widget

    def clean_up(self):
        """
        A tab gets closed on a window
        If it was the last tab in the window
            If the tab widget is the last tab widget in the main window
        destroy the tab widget
        """
        if self.count():
            return

        splitter_widget = self.splitter

        if splitter_widget.count() == 2:
            # Remove the tab widget
            self.setParent(None)
            self.hide()
            self.deleteLater()

            grandparent_splitter_widget = splitter_widget.parent()
            if not isinstance(grandparent_splitter_widget, RecursiveSplitter):
                raise RuntimeError

            widget = splitter_widget.widget(0)
            grandparent_splitter_widget.replaceWidget(
                grandparent_splitter_widget.indexOf(splitter_widget), widget
            )

        elif splitter_widget.count() == 1:
            parent = splitter_widget.parent()
            if isinstance(parent, main_window.AmuletMainWindow):
                # If this widget is the last TabEngineStackedTabWidget in the AmuletMainWindow and has no tabs, open the default tab
                # TODO: add the default page
                print("add page")
            elif isinstance(parent, sub_window.AmuletSubWindow):
                parent.deleteLater()
            else:
                raise RuntimeError
        else:
            raise RuntimeError

        # if isinstance(tab_widget, TabEngineStackedTabWidget):
        #     if not tab_widget.count():
        #         # If there are no tabs in the stacked tab widget
        #         splitter_widget = tab_widget.splitter
        #         if isinstance(splitter_widget.parent(), main_window.AmuletMainWindow) and splitter_widget.count() == 1:
        #             # If this widget is the last TabEngineStackedTabWidget in the AmuletMainWindow and has no tabs, open the default tab
        #             # TODO: add the default page
        #             print("add page")
        #         else:
        #             tab_widget.setParent(None)
        #             tab_widget.hide()
        #             tab_widget.deleteLater()
        #             clean_up_widgets(splitter_widget)
        # elif isinstance(tab_widget, RecursiveSplitter):
        #     parent = tab_widget.parent()
        #     if tab_widget.count() == 0:
        #         if isinstance(parent, sub_window.AmuletSubWindow):
        #             # widget.parent().hide()
        #             parent.deleteLater()
        #         else:
        #             raise RuntimeError
        #     elif tab_widget.count() == 1 and isinstance(parent, RecursiveSplitter):
        #         # If the widget only has one item in it them move the item into the parent splitter
        #         tab_widget = tab_widget.widget(0)
        #         if not isinstance(tab_widget, TabEngineStackedTabWidget):
        #             raise RuntimeError
        #         index = parent.indexOf(tab_widget)
        #         parent.replaceWidget(index, tab_widget)
        #         tab_widget.setParent(None)
        #         tab_widget.hide()
        #         tab_widget.deleteLater()

    def count(self) -> int:
        return self.tab_bar.count()

    def current_index(self) -> int:
        return self.tab_bar.current_index()

    @property
    def splitter(self) -> RecursiveSplitter:
        parent = self.parent()
        if not isinstance(parent, RecursiveSplitter):
            raise RuntimeError(
                "Parent of TabEngineStackedTabWidget must be RecursiveSplitter"
            )
        return parent


class RecursiveSplitter(QSplitter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setChildrenCollapsible(False)

    def addWidget(self, widget: Union[TabEngineStackedTabWidget, RecursiveSplitter]):
        if not isinstance(widget, (TabEngineStackedTabWidget, RecursiveSplitter)):
            raise TypeError(
                "widget must be an instance of TabArea or RecursiveSplitter"
            )
        super().addWidget(widget)

    def insertWidget(
        self, index: int, widget: Union[TabEngineStackedTabWidget, RecursiveSplitter]
    ):
        if not isinstance(widget, (TabEngineStackedTabWidget, RecursiveSplitter)):
            raise TypeError(
                "widget must be an instance of TabArea or RecursiveSplitter"
            )
        super().insertWidget(index, widget)



import errno
import os

import amulet_editor


def get_resource(*rel_path: str):
    path = os.path.join(amulet_editor.__path__[0], "resources", *rel_path)
    if os.path.exists(path):
        return os.path.realpath(path)
    else:
        raise FileNotFoundError(
            errno.ENOENT, "Could not find resource", os.sep.join(rel_path)
        )



import webbrowser
from dataclasses import dataclass
from functools import partial

from amulet_editor.data import build
from amulet_editor.models.widgets import ALinkCard
from PySide6.QtCore import QCoreApplication, Qt
from PySide6.QtWidgets import QLabel, QVBoxLayout, QWidget


@dataclass
class LinkData:
    name: str
    icon: str
    url: str


class StartupPanel(QWidget):
    def __init__(self):
        super().__init__()

        self.wgt_links = QWidget(self)
        self.wgt_links.setProperty("backgroundColor", "background")
        self.wgt_links.setProperty("border", "surface")
        self.wgt_links.setProperty("borderRadiusVisible", True)

        self.lbl_links = QLabel(self.wgt_links)
        self.lbl_links.setProperty("color", "on_primary")

        self.lyt_links = QVBoxLayout(self.wgt_links)
        self.lyt_links.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.lyt_links.setSpacing(5)

        self.setLayout(QVBoxLayout(self))
        self.layout().setAlignment(Qt.AlignmentFlag.AlignTop)
        self.layout().setSpacing(0)
        self.layout().addWidget(self.wgt_links)

        links: list[LinkData] = [
            LinkData(
                "Website",
                "world.svg",
                "https://www.amuletmc.com/",
            ),
            LinkData(
                "GitHub",
                "brand-github.svg",
                "https://github.com/Amulet-Team/Amulet-Map-Editor",
            ),
            LinkData(
                "Discord",
                "brand-discord.svg",
                "https://www.amuletmc.com/discord",
            ),
            LinkData(
                "Feedback",
                "flag.svg",
                "https://github.com/Amulet-Team/Amulet-Map-Editor/issues/new/choose",
            ),
        ]

        self.wgt_links.layout().addWidget(self.lbl_links)
        for link in links:
            link_card = ALinkCard(
                link.name,
                build.get_resource(f"icons/tabler/{link.icon}"),
                self.wgt_links,
            )
            link_card.clicked.connect(partial(webbrowser.open, link.url))
            self.wgt_links.layout().addWidget(link_card)

        self.wgt_links.setMaximumHeight(self.wgt_links.minimumSizeHint().height())

        self.retranslateUi()

    def retranslateUi(self):
        # Disable formatting to condense tranlate functions
        # fmt: off
        self.lbl_links.setText(QCoreApplication.translate("StartupPanel", "External Links", None))
        # fmt: on



"""A helper module to create singleton signals"""

from PySide6.QtCore import (
    Signal as _Signal,
    SignalInstance as _SignalInstance,
    QObject as _QObject,
)


def SingletonSignal(*args) -> tuple[_QObject, _SignalInstance]:
    """
    Create a singleton signal.

    :param args: The argument types the signal will have
    :return: The QObject the signal is bound to and the signal instance
    """

    class ObjCls(_QObject):
        signal = _Signal(*args)

    obj = ObjCls()
    return obj, obj.signal



from PySide6.QtCore import Qt, QCoreApplication
from PySide6.QtGui import QPixmap

from amulet_editor.data.build import get_resource

from ._splash import Ui_Splash


class Splash(Ui_Splash):
    def __init__(self):
        super().__init__()
        self.setWindowFlag(Qt.WindowType.FramelessWindowHint)
        self._logo.setPixmap(QPixmap(get_resource("icons/amulet/Icon.ico")))
        self.setWindowTitle("Amulet Editor")

    def showMessage(self, msg: str):
        self._msg.setText(msg)
        QCoreApplication.processEvents()



from __future__ import annotations

from typing import Callable, Optional
from weakref import ref
import random

from PySide6.QtCore import QCoreApplication, QMetaObject, QEvent, Qt
from PySide6.QtGui import QIcon, QShortcut
from PySide6.QtWidgets import QMainWindow, QLabel

from amulet_team_inspector import show_inspector

from amulet_team_main_window2.application.widgets.tab_engine._tab_engine import (
    TabEngineStackedTabWidget,
    RecursiveSplitter,
    TabPage,
)


class CustomTabPage(QLabel, TabPage):
    def __init__(self):
        self._name = "Tab " + str(random.randint(1, 1000))
        super().__init__(self._name)

    @property
    def name(self) -> str:
        return self._name

    @property
    def icon(self) -> Optional[QIcon]:
        return None


class AmuletMainWindow(QMainWindow):
    @staticmethod
    def _main_window() -> Optional[AmuletMainWindow]:
        return None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if AmuletMainWindow._main_window() is not None:
            raise RuntimeError("An instance of AmuletMainWindow already exists")
        AmuletMainWindow._main_window = ref(self)

        if not self.objectName():
            self.setObjectName("AmuletMainWindow")
        self.resize(800, 600)

        self.splitter_widget = RecursiveSplitter()
        tw = TabEngineStackedTabWidget()
        for i in range(1, 5):
            tw.add_page(CustomTabPage())
        self.splitter_widget.addWidget(tw)
        # tw = TabEngineStackedTabWidget()
        # for i in range(10, 20):
        #     tw.add_page(CustomTabPage())
        # self.splitter_widget.addWidget(tw)
        self.setCentralWidget(self.splitter_widget)

        self.localise()
        QMetaObject.connectSlotsByName(self)

        f12 = QShortcut(Qt.Key.Key_F12, self)
        f12.activated.connect(show_inspector)

    @classmethod
    def main_window(cls) -> AmuletMainWindow:
        main_window: Optional[AmuletMainWindow] = cls._main_window()
        if main_window is None:
            raise RuntimeError("AmuletMainWindow instance does not exist.")
        return main_window

    def changeEvent(self, event: QEvent):
        super().changeEvent(event)
        if event.type() == QEvent.Type.LanguageChange:
            self.localise()

    def localise(self):
        self.setWindowTitle(
            QCoreApplication.translate("AmuletMainWindow", "MainWindow", None)
        )



#!/usr/bin/env python3


def main():
    try:
        # Verify the python version
        import sys

        if sys.version_info[:2] < (3, 9):
            raise Exception("Must be using Python 3.9+")

        # This is required when running from a frozen bundle (eg pyinstaller)
        from multiprocessing import freeze_support

        freeze_support()

        # Initialise default paths
        from amulet_editor.data.paths._application import _init_paths

        _init_paths(None, None, None, None)

        # Import and boot the app.
        from amulet_editor.application._main import app_main

        app_main()

    except Exception as e:
        # Something crashed

        # Log the output to a file
        try:
            # If the logging module has been initialised then log to that
            import logging
            import sys
            import os
            from datetime import datetime

            if not any(
                isinstance(h, logging.FileHandler) for h in logging.getLogger().handlers
            ):
                # Set up a file handler if one does not exist

                try:
                    from amulet_editor.data.paths._application import logging_directory

                    log_dir = logging_directory()
                except Exception:
                    log_dir = "."

                file_path = os.path.join(
                    log_dir,
                    f"amulet-log-{datetime.now().strftime('%Y-%m-%d-%H-%M-%S')}-{os.getpid()}.txt",
                )
                file_handler = logging.FileHandler(file_path)

                logging.basicConfig(
                    level=logging.WARNING,
                    format="%(levelname)s - %(threadName)s - %(name)s - %(message)s",
                    force=True,
                    handlers=[logging.StreamHandler(sys.__stderr__), file_handler],
                )

            logging.exception(e)
            logging.error(
                "Amulet Crashed. Sorry about that. Please report it to a developer if you think this is an issue."
            )

        except Exception:
            # If using the logging module failed then try just writing it.
            try:
                import traceback

                tb = traceback.format_exc()
            except Exception:
                tb = None

            try:
                print(e)
                if tb is not None:
                    print(tb)
                print(
                    "Amulet Crashed. Sorry about that. Please report it to a developer if you think this is an issue."
                )
                import os

                try:
                    from amulet_editor.data.paths._application import logging_directory

                    log_dir = logging_directory()
                except Exception:
                    log_dir = "."
                with open(os.path.join(log_dir, f"crash_{os.getpid()}.log"), "w") as f:
                    if tb is not None:
                        f.write(tb)
                    f.write(str(e))
                    f.write(
                        "Amulet Crashed. Sorry about that. Please report it to a developer if you think this is an issue."
                    )
            except Exception:
                pass

        try:
            # Try reporting the crash with a GUI
            import traceback
            from PySide6.QtWidgets import QApplication
            from amulet_editor.models.widgets.traceback_dialog import (
                display_exception_blocking,
            )

            if QApplication.instance() is None:
                # QDialog needs an app otherwise it crashes
                app = QApplication()
            display_exception_blocking(
                title="Error Initialising Application",
                error=str(e),
                traceback="".join(traceback.format_exc()),
            )
        except Exception:
            pass

        try:
            # If a console is being shown then pause to let the user read it.
            import sys

            if sys.stdin is not None:
                input("Press ENTER to continue.")
        except Exception:
            pass

        raise e


if __name__ == "__main__":
    main()



from PySide6.QtCore import QLocale, QObject, Signal, QTranslator, QCoreApplication


class LocaleObject(QObject):
    locale_changed = Signal(QLocale)


_obj = LocaleObject()
locale_changed = _obj.locale_changed


def set_locale(locale: QLocale):
    """Set the application locale"""
    QLocale.setDefault(locale)
    locale_changed.emit(locale)
    # Force a language change. The above should have only modified existing translators
    t = QTranslator()
    QCoreApplication.installTranslator(t)
    QCoreApplication.removeTranslator(t)



from amulet_editor.models.package import AmuletTool, AmuletView
from amulet_editor.tools.packages._pages import PackagesPage
from amulet_editor.tools.packages._panels import PackagesPanel


class Packages(AmuletTool):
    def __init__(self):
        self._page = AmuletView(PackagesPage())
        self._primary_panel = AmuletView(PackagesPanel())
        self._secondary_panel = AmuletView(None)

    @property
    def page(self) -> AmuletView:
        return self._page

    @property
    def primary_panel(self) -> AmuletView:
        return self._primary_panel

    @property
    def secondary_panel(self) -> AmuletView:
        return self._secondary_panel

    @property
    def name(self) -> str:
        return "Packages"

    @property
    def icon_name(self) -> str:
        return "package.svg"



def enable_trace():
    """
    Enable debugging support.
    This must be called to allow debugging python function run in a QThreadPool.
    """
    try:
        # This enables debugging in PyCharm
        from _pydev_bundle.pydev_monkey_qt import set_trace_in_qt

        set_trace_in_qt()
    except Exception:
        pass



from typing import Optional
from functools import partial

from amulet_editor.data import packages, project
from amulet_editor.models.package import AmuletTool, AmuletView
from amulet_editor.models.widgets import QMenuWidget
from amulet_editor.tools.project import Project
from amulet_editor.tools.startup._models import Menu, Navigate
from amulet_editor.tools.startup.pages._new_project import NewProjectMenu
from amulet_editor.tools.startup.pages._open_world import OpenWorldMenu
from amulet_editor.tools.startup.pages._startup import StartupPage
from amulet_editor.tools.startup.panels._startup import StartupPanel
from PySide6.QtCore import QCoreApplication, QObject
from PySide6.QtWidgets import QWidget


class Startup(AmuletTool):
    def __init__(self):
        self._page = AmuletView()
        self._primary_panel = AmuletView()
        self._secondary_panel = AmuletView()

        self.view_manager = StartupManager(self)

    def set_page(self, widget: QWidget):
        self._page.setWidget(widget)

    def set_primary_panel(self, widget: QWidget):
        self._primary_panel.setWidget(widget)

    def set_secondary_panel(self, widget: Optional[QWidget]):
        self._secondary_panel.setWidget(widget)

    @property
    def page(self) -> AmuletView:
        return self._page

    @property
    def primary_panel(self) -> AmuletView:
        return self._primary_panel

    @property
    def secondary_panel(self) -> AmuletView:
        return self._secondary_panel

    @property
    def name(self) -> str:
        return "Startup"

    @property
    def icon_name(self) -> str:
        return "hexagons.svg"


class StartupManager(QObject):
    def __init__(self, plugin: Startup):
        super().__init__()

        self.menu_list: list[Menu] = []
        self.plugin = plugin

        self.startup_page = StartupPage()
        self.startup_panel = StartupPanel()
        self.menu_page = QMenuWidget()

        # Connect signals
        self.startup_page.crd_open_level.clicked.connect(
            partial(self.set_menu_page, OpenWorldMenu)
        )
        self.startup_page.crd_new_project.clicked.connect(
            partial(self.set_menu_page, NewProjectMenu)
        )

        self.menu_page.clicked_cancel.connect(partial(self.set_startup_page))
        self.menu_page.clicked_back.connect(partial(self.menu_back))
        self.menu_page.clicked_next.connect(partial(self.menu_next))

        self.set_startup_page()

    def set_startup_page(self):
        self.close_menu()

        # Set plugin view
        self.plugin.set_page(self.startup_page)
        self.plugin.set_primary_panel(self.startup_panel)
        self.plugin.set_secondary_panel(None)

    def set_menu_page(self, menu_class):
        # Create menu
        menu: Menu = menu_class(self.plugin.set_secondary_panel)
        menu.enable_next.connect(self.menu_page.btn_next.setEnabled)

        self.menu_list.append(menu)
        self.set_menu(menu)

        # Set plugin view
        self.plugin.set_page(self.menu_page)

    def set_menu(self, menu: Menu, new: bool = True):
        # Update menu page
        menu_page = self.menu_page
        menu_page.btn_back.setVisible(menu is not self.menu_list[0])
        menu_page.btn_next.setEnabled(False)
        menu_page.setMenuTitle(menu.title)
        menu_page.setWidget(menu.widget())

        if menu.next_menu() is None:
            menu_page.btn_next.setText(
                QCoreApplication.translate("QMenuWidget", "Finish", None)
            )
        else:
            menu_page.btn_next.setText(
                QCoreApplication.translate("QMenuWidget", "Next", None)
            )

        # Set plugin view
        menu.navigated(Navigate.HERE)
        self.plugin.set_secondary_panel(None)

    def close_menu(self):
        for menu in self.menu_list:
            self.menu_page.removeWidget(menu.widget())
            menu.widget().deleteLater()

        self.menu_list = []

    def menu_back(self):
        current_menu = self.menu_list[-1]
        current_menu.navigated(Navigate.BACK)

        self.menu_list.remove(current_menu)
        self.set_menu(self.menu_list[-1], False)

    def menu_next(self):
        current_menu = self.menu_list[-1]
        current_menu.navigated(Navigate.NEXT)

        if current_menu.next_menu() is not None:
            menu: Menu = current_menu.next_menu()
            menu.enable_next.connect(self.menu_page.btn_next.setEnabled)

            self.menu_list.append(menu)
            self.set_menu(self.menu_list[-1])

    def set_new_project_page(self):
        page = self.new_project_page

        def enable_next(project_name: str):
            project_name = project_name.strip()
            page.btn_next.setEnabled(len(project_name) > 0)

        # Connect signals
        page.lne_project_name.textChanged.connect(enable_next)
        page.btn_cancel.clicked.connect(partial(self.set_startup_page))
        page.btn_next.clicked.connect(partial(self.set_import_level_page))

        # Set plugin view
        self.plugin.set_secondary_panel(None)

    def set_import_level_page(self):
        page = self.import_level_page

        # Connect signals
        page.btn_cancel.clicked.connect(partial(self.set_startup_page))
        page.btn_back.clicked.connect(partial(self.set_new_project_page))

        # Set plugin view
        self.plugin.set_secondary_panel(None)

    def set_select_packages_page(self):
        page = self.select_packages_page

        def open_project():
            tools = packages.list_tools()
            for tool in tools:
                packages.disable_tool(tool)

            packages.enable_tool(Project())
            project.set_root(self.project_root)

        # Connect signals
        page.btn_cancel.clicked.connect(partial(self.set_startup_page))
        page.btn_back.clicked.connect(partial(self.set_open_world_page))
        page.btn_next.clicked.connect(partial(open_project))

        # Set plugin view
        self.plugin.set_secondary_panel(None)



from PyInstaller.utils.hooks import (
    collect_data_files,
    copy_metadata,
    collect_submodules,
)

datas = [
    *collect_data_files("amulet_editor", excludes=["**/*.ui", "**/*.c", "**/*.pyx"]),
    *collect_data_files(
        "amulet_editor.plugins",
        include_py_files=True,
        excludes=["**/*.ui", "**/*.c", "**/*.pyx", "**/*.pyc"],
    ),
    *copy_metadata("amulet_editor", recursive=True),
]

hiddenimports = [
    *collect_submodules("amulet_editor", lambda name: name != "amulet_editor.plugins"),
    "PySide6",
    "OpenGL"
]



from ._card import ALinkCard, QPixCard
from ._draggable import ADragContainer
from ._icon import AStylableSvgWidget, AIconButton, ATooltipIconButton
from ._label import QElidedLabel
from ._menu import QMenuWidget
from ._line import QHLine



from amulet_editor.models.package import AmuletTool, AmuletView
from amulet_editor.tools.project.pages._project import ProjectPage
from amulet_editor.tools.project.panels._project import ProjectPanel
from PySide6.QtCore import QObject
from PySide6.QtWidgets import QWidget


class Project(AmuletTool):
    def __init__(self):
        self._page = AmuletView()
        self._primary_panel = AmuletView()
        self._secondary_panel = AmuletView()

        self.view_manager = ProjectManager(self)

    def set_page(self, widget: QWidget):
        self._page.setWidget(widget)

    def set_primary_panel(self, widget: QWidget):
        self._primary_panel.setWidget(widget)

    def set_secondary_panel(self, widget: QWidget):
        self._secondary_panel.setWidget(widget)

    @property
    def page(self) -> AmuletView:
        return self._page

    @property
    def primary_panel(self) -> AmuletView:
        return self._primary_panel

    @property
    def secondary_panel(self) -> AmuletView:
        return self._secondary_panel

    @property
    def name(self) -> str:
        return "Project"

    @property
    def icon_name(self) -> str:
        return "folders.svg"


class ProjectManager(QObject):
    def __init__(self, plugin: Project):
        super().__init__()

        self.plugin = plugin

        self.project_page = ProjectPage()

        self.project_panel = ProjectPanel()

        self.set_project_page()
        self.set_project_panel()

    def set_project_page(self):
        page = self.project_page

        self.plugin.set_page(page)

    def set_project_panel(self):
        panel = self.project_panel

        panel.file_selected.connect(self.project_page.show_file)

        self.plugin.set_primary_panel(panel)



from __future__ import annotations
from typing import Optional

from .application.windows.main_window import AmuletMainWindow


window: Optional[AmuletMainWindow] = None


def load_plugin():
    global window
    window = AmuletMainWindow()
    window.showMaximized()
